{
 DX8DM ---- Pulsar DirectX 8 DirectMusic unit.
            by Lord Trancos.

            Currently, designed to be the sound-core only.
}

unit DX8DM;

// *******************************************************************

interface

uses Windows, ActiveX, DirectMusic;

const
  DMMINAUDIOVOLUME = -9600; // Minimum audio volume.

var
  DM8Loader      : IDirectMusicLoader8 = NIL;      // File loader.
  DM8Performance : IDirectMusicPerformance8 = NIL; // The player.
  DM8AudioPath   : IDirectMusicAudioPath8 = NIL;   // Audio path.

function  DMInit(_hWnd: hWnd; _reverb: boolean): boolean;
procedure DMClose;

function  DMLoadWavFile(_filename: string): IDirectMusicSegment8;
function  DMPlay(_segment: IDirectMusicSegment8): boolean;
function  DMStop(_segment: IDirectMusicSegment8): boolean;
procedure DMFreeSegment(var _segment: IDirectMusicSegment8);

// *******************************************************************

implementation

const
  DM8PERFORMANCE_CHANNELS = 16; // # of performance channels.

// *******************************************************************
// Name : DMInit.
// Desc : Creates Loader and Performance Interfaces.
//        Also Initializes audio engine.
// *******************************************************************

function DMInit(_hWnd: hWnd; _reverb: boolean): boolean;

var
  _guid : TGUID;
  _path : integer;

begin
  Result := false;

  // Initialize COM.
  if failed(CoInitialize(NIL)) then exit;

  // Create IDirectMusicLoader8 Interface.
  Move(IID_IDirectMusicLoader8, _guid, SizeOf(TGUID));
  if failed(CoCreateInstance(TGUID(CLSID_DirectMusicLoader),
            NIL, CLSCTX_INPROC, _guid, DM8Loader)) then exit;

  // Create IDirectMusicPerformance8 Interface.
  Move(IID_IDirectMusicPerformance8, _guid, SizeOf(TGUID));
  if failed(CoCreateInstance(TGUID(CLSID_DirectMusicPerformance),
            NIL, CLSCTX_INPROC, _guid, DM8Performance)) then exit;

  // Initialize the performance and the synthesizer.
  if _reverb
    then _path := DMUS_APATH_SHARED_STEREOPLUSREVERB
      else _path := DMUS_APATH_DYNAMIC_STEREO;
  if failed(DM8Performance.InitAudio(NIL, NIL, _hWnd, _path,
            DM8PERFORMANCE_CHANNELS, DMUS_AUDIOF_ALL, NIL)) then exit;

  // Retrieve audiopath set by InitAudio.
  if failed(DM8Performance.GetDefaultAudioPath(DM8AudioPath)) then exit;

  // All right.
  Result := true;
end;

// *******************************************************************
// Name : DMClose.
// Desc : Close all. Before, you must release all segments.
// *******************************************************************

procedure DMClose;

begin
  // Close down the performance
  if assigned(DM8Performance) then DM8Performance.CloseDown;

  // Release all interfaces
  if assigned(DM8AudioPath) then DM8AudioPath := NIL;
  if assigned(DM8Performance) then DM8Performance := NIL;
  if assigned(DM8Loader) then DM8Loader := NIL;

  // Close COM
  CoUninitialize;
end;

// *******************************************************************
// Name : DMLoadFile.
// Desc : Loads a file into a Segment. (usually a WAV file.)
// *******************************************************************

function DMLoadWavFile(_filename: string): IDirectMusicSegment8;

var _wFileName : array[0..MAX_PATH-1] of WCHAR;

begin
  Result := NIL;

  // Get filename in UNICODE
  MultiByteToWideChar(CP_ACP, 0, pChar(_filename), -1, @_wFileName,
                      MAX_PATH);

  // Load file.
  if failed(DM8Loader.LoadObjectFromFile(
            CLSID_DirectMusicSegment,   // Class identifier.
            IID_IDirectMusicSegment8,   // ID of desired interface.
            _wFileName,                 // Filename.
            Result // Pointer that receives interface.
            )) then begin Result := NIL; exit; end;

  // Downloads the band to the performance.
  if failed(Result.Download(DM8Performance)) then
    begin Result := NIL; exit; end;

  // All right.
end;

// *******************************************************************
// Name : DMPlay.
// Desc : Plays a segment.
// *******************************************************************

function DMPlay(_segment: IDirectMusicSegment8): boolean;

var _p: IDirectMusicSegmentState;

begin
  Result := false;
  if not (assigned(_segment)) then exit;

  _p := NIL;
  Result := not failed(DM8Performance.PlaySegment(
                       _segment,  // Segment to play.
                       DMUS_SEGF_SECONDARY, // Flags.
                       0,   // Start time; 0 is immediate.
                       @_p  // Pointer that receives segment state.
                       ));
end;

// *******************************************************************
// Name : DMStop.
// Desc : Stops a segment.
// *******************************************************************

function DMStop(_segment: IDirectMusicSegment8): boolean;

begin
  Result := false;
  if not (assigned(_segment)) then exit;

  Result := not failed(DM8Performance.Stop(
                       _segment, // Stop all segments.
                       NIL,      // Stop all segment states.
                       0,        // Do it immediately.
                       0         // Flags.
                       ));
end;

// *******************************************************************
// Name : DMFreeSegment.
// Desc : Frees a segment.
// *******************************************************************

procedure DMFreeSegment(var _segment: IDirectMusicSegment8);

begin
  if assigned(_segment) then
    begin _segment.Unload(NIL); _segment := NIL; end;
end;

// *******************************************************************

end.

