{
 DX8DSMP -- Pulsar DirectX 8 DirectShow Media Player unit.
            by Lord Trancos.

            Currently, designed to be the music-core only.

 2000/12/03 - First release with DSPlayer.pas name.
 2000/12/16 - Rewritten, all changed.
}

unit DX8DSMP;

// ------------------------------------------------------------

interface

uses Windows, ActiveX, DirectShow;

const
  DSMINAUDIOVOLUME = -10000; // Minimum audio volume.

type
  TDSMP = record
    Initialized  : boolean;
    // DirectShow stuff.
    GraphBuilder : IGraphBuilder;
    MediaControl : IMediaControl;
    MediaSeeking : IMediaSeeking;
    MediaEventEx : IMediaEvent;
    BasicAudio   : IBasicAudio;
    BasicVideo   : IBasicVideo;
    VideoWindow  : IVideoWindow;
    // About the media file.
    AudioAvail   : boolean;
    VideoAvail   : boolean;
    ValidFile    : boolean;
    // Video info.
    VideoWidth   : integer;
    VideoHeight  : integer;
    VideoBitRate : integer;
  end;

procedure dsmpReset(var _dsmp: TDSMP);
function  dsmpLoadFile(var _dsmp: TDSMP; _filename: string): boolean;
procedure dsmpFree(var _dsmp: TDSMP);                            

// ------------------------------------------------------------

implementation

// ------------------------------------------------------------

procedure dsmpReset(var _dsmp: TDSMP);

begin
  FillChar(_dsmp, SizeOf(_dsmp), 0);
end;

// ------------------------------------------------------------

procedure dsmpFree(var _dsmp: TDSMP);

begin
  with _dsmp do
    begin
      if not Initialized then exit;

      // Hide VideoWindow and relinquish ownership.
      if Assigned(VideoWindow) then
        begin
          VideoWindow.put_Visible(false);
          VideoWindow.put_Owner(0);
        end;

      if Assigned(MediaEventEx) then MediaEventEx := NIL;
      if Assigned(MediaSeeking) then MediaSeeking := NIL;
      if Assigned(MediaControl) then MediaControl := NIL;
      if Assigned(BasicAudio)   then BasicAudio   := NIL;
      if Assigned(BasicVideo)   then BasicVideo   := NIL;
      if Assigned(VideoWindow)  then VideoWindow  := NIL;
      if Assigned(GraphBuilder) then GraphBuilder := NIL;

      dsmpReset(_dsmp);
    end;
end;

// ------------------------------------------------------------

function dsmpLoadFile(var _dsmp: TDSMP; _filename: string): boolean;

var _wfn : array[0..MAX_PATH-1] of wchar;
    _vis : longBool;

begin
  Result := false;

  // Free previous interfaces.
  dsmpFree(_dsmp);

  // Get unicode filename.
  MultiByteToWideChar(CP_ACP, 0, pChar(_filename), -1, @_wfn, MAX_PATH);

  // Create interfaces and construct the graph.
  with _dsmp do
    begin
      // Initialized.
      Initialized := true;

      // Create DirectShow Graph.
      if failed(CoCreateInstance(TGUID(CLSID_FilterGraph),
                                 NIL, CLSCTX_INPROC_SERVER,
                                 TGUID(IID_IGraphBuilder),
                                 GraphBuilder)) then exit;

      // Build the filter graph.
      if failed(GraphBuilder.RenderFile(_wfn, NIL)) then exit;

      // Get the IMediaControl Interface
      if failed(GraphBuilder.QueryInterface(IID_IMediaControl,
                MediaControl)) then exit;

      // Get the IMediaSeeking Interface
      if failed(GraphBuilder.QueryInterface(IID_IMediaSeeking,
                MediaSeeking)) then exit;

      // Get the IMediaEventEx Interface
      if failed(GraphBuilder.QueryInterface(IID_IMediaEventEx,
                MediaEventEx)) then exit;

      // Get Audio and Video Interfaces.
      GraphBuilder.QueryInterface(IID_IBasicAudio,  BasicAudio);
      GraphBuilder.QueryInterface(IID_IBasicVideo,  BasicVideo);
      GraphBuilder.QueryInterface(IID_IVideoWindow, VideoWindow);

      // Get file info.
      AudioAvail := Assigned(BasicAudio);
      VideoAvail := Assigned(BasicVideo) and Assigned(VideoWindow) and
                    (not failed(VideoWindow.get_Visible(_vis)));
      ValidFile  := (AudioAvail) or (VideoAvail);

      // Get video info.
      if VideoAvail then
        begin
          BasicVideo.GetVideoSize(VideoWidth, VideoHeight);
          BasicVideo.get_BitRate(VideoBitRate);
        end;
    end;

  Result := true;
end;

// ------------------------------------------------------------

end.
