{
 DXDraw --- DirectDraw unit.
            Requieres DirectDraw.pas from 
            DirectX 8.0 headers for Delphi.

            You can find those headers here;
            http://www.delphi-jedi.org/DelphiGraphics/
}

unit DXDraw;

// *******************************************************************

interface

uses Windows, SysUtils, Classes, DirectDraw;

var
  DD7 : IDirectDraw7; // DirectDraw Interface.
  DDMAINS : IDirectDrawSurface7; // Primary surface.
  DDBACKS : IDirectDrawSurface7; // Backbuffer surface.
  DDVIEWPORT  : TRect;   // Viewport.
  DD_BFWIDTH  : longint; // Width and height of
  DD_BFHEIGHT : longint; // the backbufer.
  DDWindowedModeSrcRect : TRect;

// Initialization and closing functions.
function  DDInit(_GUID: pGUID; _hWnd: HWND;
                 _fullScr: boolean): boolean;
function  DDSetDisplayMode(_width, _height, _bpp,
                           _refreshRate: longint): boolean;
function  DDCreateFullScrBuffers: boolean;
function  DDCreateWindowBuffers(_hWnd: HWND): boolean;
procedure DDFreeMainBuffers;
procedure DDClose;

// Othe functions.
procedure DDFlip(_hWnd: hWnd);
function  DDSetClipper(_hWnd: hWnd;
                       _surf: IDirectDrawSurface7): boolean;
//function  DDRestoreMainBuffers: boolean;                       

// *******************************************************************

implementation

var DDAreInFullScreenMode : boolean;

// *******************************************************************
// Name : DDFreeMainBuffers.
// Desc : Free DDMAINS and DDBACKS.
// *******************************************************************

procedure DDFreeMainBuffers;

begin
  if Assigned(DDBACKS) then DDBACKS := NIL;
  if Assigned(DDMAINS) then DDMAINS := NIL;
end;

// *******************************************************************
// Name : DDClose.
// Desc : Closes DirectDraw.
// *******************************************************************

procedure DDClose;

begin
  DDFreeMainBuffers;
  if Assigned(DD7) then DD7 := NIL;
  DD_BFWIDTH  := 0;
  DD_BFHEIGHT := 0;
end;

// *******************************************************************
// Name : DDInit.
// Desc : Initialize DirectDraw.
// *******************************************************************

function DDInit(_GUID: pGUID; _hWnd: HWND;
                _fullScr: boolean): boolean;

var _res  : hResult;
    _coop : cardinal;

begin
  Result := false;

  // Close DirectDraw (if was already initialized).
  DDClose;

  // Set some global variables.
  DDAreInFullScreenMode  := _fullScr;
  GetClientRect(_hWnd, DDVIEWPORT);
  GetClientRect(_hWnd, DDWindowedModeSrcRect);

  // Create DirectDraw Interface.
  _res := DirectDrawCreateEx(_GUID, DD7, IID_IDirectDraw7, nil);
  if failed(_res) then exit;

  // Set cooperation level.
  if not _fullScr
    then _coop := DDSCL_NORMAL
      else _coop := DDSCL_ALLOWREBOOT or DDSCL_EXCLUSIVE or
                    DDSCL_FULLSCREEN;
  _coop := _coop or DDSCL_FPUSETUP;
  _res  := DD7.SetCooperativeLevel(_hWnd, _coop);

  // Return result.
  Result := not failed(_res);
end;

// *******************************************************************
// Name : DDSetDisplayMode.
// Desc : Change the display mode (fullscreen).
// *******************************************************************

function DDSetDisplayMode(_width, _height, _bpp,
                          _refreshRate: longint): boolean;

var _mode : cardinal;
     _res : hResult;

begin
  Result := false;

  // If the selected mode is 320x200x8, use the 13h mode.
  // (do not use a X-mode)
  if (_width = 320) and (_height = 200) and (_bpp = 8)
    then _mode := DDSDM_STANDARDVGAMODE
      else _mode := 0;

  // Set the video mode.
  _res := DD7.SetDisplayMode(_width, _height, _bpp,
                             _refreshrate, _mode);

  // All right ?
  if not failed(_res) then
    begin
      DDVIEWPORT  := Rect(0, 0, _width, _height);
      DD_BFWIDTH  := _width;
      DD_BFHEIGHT := _height;
      Result  := true;
    end;
end;

// *******************************************************************
// Name : DDCreateFullScrBuffers.
// Desc : Initializes DDMAINS and DDBACKS for fullscreen mode.
// *******************************************************************

function DDCreateFullScrBuffers: boolean;

var _res : hResult;
    _ddsd : TDDSurfaceDesc2;
    _ddsCaps : TDDSCaps2;

begin
  Result := false;

  // Free the buffers.
  DDFreeMainBuffers;

  // Set the primary surface description.
  ZeroMemory(@_ddsd, SizeOf(_ddsd));
  with _ddsd do
    begin
      dwSize  := SizeOf(_ddsd);
      dwFlags := DDSD_CAPS or DDSD_BACKBUFFERCOUNT;
      ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or DDSCAPS_3DDEVICE or
                        DDSCAPS_FLIP or DDSCAPS_COMPLEX;
      dwBackBufferCount := 1;
    end;

  // Create the primary surface.
  _res := DD7.CreateSurface(_ddsd, DDMAINS, nil);
  if failed(_res) then exit;

  // Get the backbuffer from the primary surface.
  ZeroMemory(@_ddsCaps, SizeOf(_ddsCaps));
  _ddsCaps.dwCaps := DDSCAPS_BACKBUFFER;
  _res := DDMAINS.GetAttachedSurface(_ddscaps, DDBACKS);
  DDBACKS._AddRef;

  // Return the result.
  Result := not failed(_res);
end;

// *******************************************************************
// Name : DDCreateWindowBuffers.
// Desc : Initializes DDMAINS and DDBACKS for windowed mode.
// *******************************************************************

function DDCreateWindowBuffers(_hWnd: HWND): boolean;

var _res : hResult;
    _rect : TRect;
    _ddsd : TDDSurfaceDesc2;
//    _clipper : IDirectDrawClipper;

begin
  Result := false;

  // Free the buffers.
  DDFreeMainBuffers;

  // Set the primary surface description.
  ZeroMemory(@_ddsd, SizeOf(_ddsd));
  with _ddsd do
    begin
      dwSize := sizeof(_ddsd);
      dwFlags := DDSD_CAPS;
      ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE;
    end;

  // Create the primary surface.
  _res := DD7.CreateSurface(_ddsd, DDMAINS, nil);
  if failed(_res) then exit;

  // Set the backbuffer description.
  with _ddsd do
    begin
      dwFlags  := DDSD_WIDTH or DDSD_HEIGHT or DDSD_CAPS;
      GetClientRect(_hWnd, _rect);
      ClientToScreen(_hWnd, _rect.TopLeft);
      ClientToScreen(_hWnd, _rect.BottomRight);
      dwWidth  := _rect.right - _rect.left;
      dwHeight := _rect.bottom - _rect.top;
      ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN or DDSCAPS_3DDEVICE;
    end;

  // Create the backbuffer.
  _res := DD7.CreateSurface(_ddsd, DDBACKS, nil);
  if failed(_res) then exit;
{
  // Create a clipper interface.
  _res := DD7.CreateClipper(0, _clipper, nil);
  if failed(_res) then exit;

  // Set the clipper.
  _clipper.SetHWnd(0, _hWnd);
  DDMAINS.SetClipper(_clipper);
}
  // All right.
  DD_BFWIDTH  := _ddsd.dwWidth;
  DD_BFHEIGHT := _ddsd.dwHeight;
  Result := true;
end;

// *******************************************************************
// Name : DDRestoreMainBuffers.
// Desc : Restores the primary and back buffers.
//        Returns true if the buffers were lost.
// *******************************************************************

function DDRestoreMainBuffers: boolean;

var _b1, _b2 : boolean;

begin
  _b1 := false;
  _b2 := false;
  if Assigned(DDBACKS) then
    begin
      _b1 := (DDBACKS.IsLost <> 0);
      DDBACKS._Restore;
    end;
  if Assigned(DDMAINS) then
    begin
      _b2 := (DDMAINS.IsLost <> 0);
      DDMAINS._Restore;
    end;
  Result := (_b1 and _b2);
end;

// *******************************************************************
// Name : DDFlip.
// Desc : Makes a page flipping between the main and back buffers.
// *******************************************************************

procedure DDFlip(_hWnd: hWnd);

//var _hr : hResult;

begin
   if not DDAreInFullScreenMode then
    begin
      GetClientRect(_hWnd, DDVIEWPORT);
      ClientToScreen(_hWnd, DDVIEWPORT.TopLeft);
      ClientToScreen(_hWnd, DDVIEWPORT.BottomRight);
      {_hr :=} DDMAINS.Blt(@DDVIEWPORT, DDBACKS,
                         @DDWindowedModeSrcRect, DDBLT_WAIT, NIL);
    end else {_hr :=} DDMAINS.Flip(DDBACKS, 0);
//  if _hr = DDERR_SURFACELOST then DDRestoreMainBuffers;
end;

// *******************************************************************
// Name : DDSetClipper.
// Desc : Sets a default clipper for FullScreen and Windowed modes.
// *******************************************************************

function DDSetClipper(_hWnd: hWnd;
                      _surf: IDirectDrawSurface7): boolean;

var _clipper : IDirectDrawClipper;
    _rgn : hRGN;
    _rgnsiz : longint;
    _rgndata : pointer;
    _hr : hResult;

begin
  Result := false;

  // Create the clipper interface.
  _hr := DD7.CreateClipper(0, _clipper, NIL);
  if failed(_hr) then exit;

  // Create Rgn
  with DDVIEWPORT do _rgn := CreateRectRgn(Left, Top, Right, Bottom);
  _rgnsiz := GetRegionData(_rgn, 0, NIL);
  if (_rgn <> 0) and (_rgnsiz > 0) then
    begin
      // Set region.
      GetMem(_rgndata, _rgnsiz);
      if GetRegionData(_rgn, _rgnsiz, _rgndata) > 1
        then _hr := _clipper.SetClipList(_rgndata, 0);
      FreeMem(_rgndata, _rgnsiz);
      // Free region.
      DeleteObject(_rgn);
    end else
    begin _clipper := NIL; exit; end;

  // if we have a valid clipper, then set it to the surface.
  if not failed(_hr) then _hr := _surf.SetClipper(_clipper);

  // Release the clipper.
  _clipper := NIL;

  // Return result.
  Result := not failed(_hr);
end;

// *******************************************************************

begin
  DD7 := NIL;
  DDMAINS := NIL;
  DDBACKS := NIL;
  DDVIEWPORT := Rect(0, 0, 0, 0);
  DD_BFWIDTH  := 0;
  DD_BFHEIGHT := 0;
  DDAreInFullScreenMode := false;
end.
