{
 DXGDI ---- DX+GDI unit.
            Requieres DirectDraw.pas from 
            DirectX 8.0 headers for Delphi.

            You can find those headers here;
            http://www.delphi-jedi.org/DelphiGraphics/
}

unit DXGDI;

// *******************************************************************

interface

uses Windows, DirectDraw;

function DDDrawBitmap(_SrcBMP: hBitmap;
                      _SrcX, _SrcY, _SrcW, _SrcH,
                      _DstX, _DstY, _DstW, _DstH: longint;
                      _DestSurf: IDirectDrawSurface7): boolean;
function DDColorMatch(_surf: IDirectDrawSurface7;
                      _rgb: TColorRef): TColorRef;

// *******************************************************************

implementation

// *******************************************************************
// Name : DDDrawBitmap.
// Desc : Draws a bitmap on a DirectDraw surface.
// *******************************************************************

function DDDrawBitmap(_SrcBMP: hBitmap;
                      _SrcX, _SrcY, _SrcW, _SrcH,
                      _DstX, _DstY, _DstW, _DstH: longint;
                      _DestSurf: IDirectDrawSurface7): boolean;

var _dcBMP  : hDC;
    _dcSurf : hDC;
    _hr : hResult;

begin
  Result := false;

  // Create DC for the BMP.
  _dcBMP := CreateCompatibleDC(0);
  if _dcBMP = 0 then exit;
  SelectObject(_dcBMP, _SrcBMP);

  // Get DC for the surface.
  _hr := _DestSurf.GetDC(_dcSurf);
  if not failed(_hr) then
    begin
      // Draw.
      Result := StretchBlt(_dcSurf, _DstX, _DstY, _DstW, _DstH, _dcBMP,
                           _SrcX, _SrcY, _SrcW, _SrcH, SRCCOPY);
      // Free surface's DC.
      _DestSurf.ReleaseDC(_dcSurf);
    end;

  // Free BMP's DC.
  DeleteDC(_dcBMP);
end;

// *******************************************************************
// Name : DDColorMatch.
// Desc : Returns the nearest color (to _rgb) in the selected surface.
//        If _rgb is CLR_INVALID the function returns the color
//        located on the 0,0 coordinates.
// Warning : This function must be used only under non paletized
//           video modes (16, 24 or 32 bpp).
// *******************************************************************

function DDColorMatch(_surf: IDirectDrawSurface7;
                      _rgb: TColorRef): TColorRef;

var _hdc  : HDC;
    _rgbT : TColorRef;
    _ddsd : TDDSurfaceDesc2;
    _val  : longint;
    _hr   : hResult;

begin
  Result := CLR_INVALID;
  _rgbT := CLR_INVALID;

  // Check the surface.
  if _surf = NIL then exit;

  // GDI's SetPixel function will do the Job for us.
  if (_rgb <> CLR_INVALID) then
    begin
      if _surf.GetDC(_hdc) <> DD_OK then exit;

      // Save the 0,0 pixel and change it.
      _rgbT := GetPixel(_hdc, 0, 0);
      SetPixel(_hdc, 0, 0, _rgb);
      _surf.ReleaseDC(_hdc);
    end;

  // Lock the surface...
  _ddsd.dwSize := sizeof(_ddsd);
  repeat
    _hr := _surf.Lock(NIL, _ddsd, 0, 0);
  until _hr <> DDERR_WASSTILLDRAWING;

  // ... and read the value of the 0,0 coordinate.
  if _hr = DD_OK then
    begin
      // Read 4 bytes.
      _val := longint(_ddsd.lpSurface^);
      // Adjust for the current bpp (Bits Per Pixel).
      if (_ddsd.ddpfPixelFormat.dwRGBBitCount < 32)
        then _val := _val and ($FFFFFFFF shr (32 - _ddsd.ddpfPixelFormat.dwRGBBitCount));
      // Unlock.
      _surf.Unlock(NIL);
      // All right.
      Result := _val;
    end;

  // Restore the color in the 0,0 coordinate.
  if (_rgb <> CLR_INVALID) and (_surf.GetDC(_hdc) = DD_OK) then
    begin
      SetPixel(_hdc, 0, 0, _rgbT);
      _surf.ReleaseDC(_hdc);
    end;
end;

// ******************************************************************

end.
