{
 DXSPRGRD - Sprite Grids Sources unit.
            Requieres DirectDraw.pas from
            DirectX 8.0 headers for Delphi.

            You can find those headers here;
            http://www.delphi-jedi.org/DelphiGraphics/
}

unit DXSprGrd;

// *******************************************************************

interface

uses Windows, Classes, DirectDraw, DXSprSrc;

type // Surface grid.
     TDDSurfGrid = record
                     Rects : pointer; // Array of TRect.
                     Count : longint; // Number of Rects.
                   end;

     // Sprite grid source.
     TDDSprGrdSrc = record
                      SprSrc : TDDSprSrc;
                      Grid   : TDDSurfGrid;
                    end;
     pTDDSprGrdSrc = ^TDDSprGrdSrc;

// Surface grid functions.
procedure DDResetSurfGrid(var _grid: TDDSurfGrid);
function  DDCreateSurfGrid(_surf: IDirectDrawSurface7;
                           var _grid: TDDSurfGrid): boolean;
procedure DDFreeSurfGrid(var _grid: TDDSurfGrid);

// Sprite grid functions.
procedure DDResetSprGrdSrc(var _sprGrdSrc: TDDSprGrdSrc);
function  DDCreateSprGrdSrc(_filename: string): TDDSprGrdSrc;
function  DDRestoreSprGrdSrc(var _sprGrdSrc: TDDSprGrdSrc): boolean;
procedure DDFreeSprGrdSrc(var _sprGrdSrc: TDDSprGrdSrc);
function  DDValidSprGrdSrc(_sprGrdSrc: TDDSprGrdSrc): boolean;

// *******************************************************************

implementation

// *******************************************************************
// Name : DDResetSurfGrid.
// Desc : Sets to 0 a TDDSurfGrid variable.
// *******************************************************************

procedure DDResetSurfGrid(var _grid: TDDSurfGrid);

begin
  _grid.Rects := NIL;
  _grid.Count := 0;
end;

// *******************************************************************
// Name : DDGetSurfaceGrid.
// Desc : Obtains an array of rects (sprite cells) from a valid
//        sprite-grid surface.
// *******************************************************************

function DDCreateSurfGrid(_surf: IDirectDrawSurface7;
                          var _grid: TDDSurfGrid): boolean;

var _hdc  : HDC;
    _gcol : TColorRef; // Grid color.
    _ccol : TColorRef; // Current color.
    _desc : TDDSurfaceDesc2; // Surface desc.
    _nrects : longint; // Number of rects. (Grid cells)
    _cursor : ^TRect;
    _srcX, _srcY : longint;
    _cntA, _cntX : longint;
    _cntY : longint;

begin
  Result := false;
  DDResetSurfGrid(_grid);

  // Check the surface.
  if _surf = NIL then exit;

  // Get surface's description.
  FillChar(_desc, SizeOf(_desc), 0);
  _desc.dwSize := SizeOf(_desc);
  _surf.GetSurfaceDesc(_desc);

  // Check dimensions.
  if (_desc.dwWidth <= 2) or (_desc.dwHeight <= 2) then exit;

  // Get a DC for the surface.
  if _surf.GetDC(_hdc) <> DD_OK then exit;

  // Get the grid's color.
  _gcol := GetPixel(_hdc, 0, 0);

  // First count the rects, after save the rects.
  _nrects := 0;
  _cursor := NIL;
  for _cntA := 0 to 1 do
    begin
      // Get memory for the buffer.
      if (_cntA = 1) then
        try
          GetMem(_grid.Rects, SizeOf(TRect) * _nrects);
          _grid.Count := _nrects;
          _cursor := _grid.Rects;
        except _surf.ReleaseDC(_hdc); exit; end;

      // Find rects...
      _cntY := 0;
      _nrects := 0;
      repeat
        inc(_cntY);
        _srcY := _cntY;

        while (_cntY < longint(_desc.dwHeight))
          and (GetPixel(_hdc, 1, _cntY) <> _gcol)
           do inc(_cntY);

        if _cntY - 1 > _srcY then
          begin
            _srcX := 1;
            for _cntX := 1 to _desc.dwWidth - 1 do
              begin
                _ccol := GetPixel(_hdc, _cntX, _cntY - 1);
                if (_ccol = _gcol) and (_cntX - 1 > _srcX) then
                  begin
                    // Save current rect.
                    if (_cntA = 1) and (_cursor <> NIL) then
                      begin
                        _cursor^ := Rect(_srcX, _srcY, _cntX, _cntY);
                        inc(longint(_cursor), SizeOf(TRect));
                        _srcX := _cntX + 1;
                      end;
                    inc(_nrects);
                  end;
              end;
          end;
      until _cntY = longint(_desc.dwHeight) - 1;
    end;

  // Release DC.
  _surf.ReleaseDC(_hdc);

  // All ok
  Result := true;
end;

// *******************************************************************
// Name : DDResetSurfGrid.
// Desc : Frees a TDDSurfGrid variable.
// *******************************************************************

procedure DDFreeSurfGrid(var _grid: TDDSurfGrid);

begin
  if _grid.Rects <> NIL then
    try FreeMem(_grid.Rects, SizeOf(TRect) * _grid.Count);
    except end;
  _grid.Rects := NIL;
  _grid.Count := 0;
end;

// *******************************************************************
// Name : DDResetSprGrd.
// Desc : Sets to 0 a TDDSprGrdSrc variable.
// *******************************************************************

procedure DDResetSprGrdSrc(var _sprGrdSrc: TDDSprGrdSrc);

begin
  DDResetSprSrc(_sprGrdSrc.SprSrc);
  DDResetSurfGrid(_sprGrdSrc.Grid);
end;

// *******************************************************************
// Name : DDCreateSprGrdSrc.
// Desc : Create a "Sprite Grid Source" from a BMP file.
// *******************************************************************

function DDCreateSprGrdSrc(_filename: string): TDDSprGrdSrc;

begin
  Result.SprSrc := DDCreateSprSrc(_filename);
  if Result.SprSrc.Data <> NIL then
    if not DDCreateSurfGrid(Result.SprSrc.Data, Result.Grid)
      then DDFreeSprSrc(Result.SprSrc);
end;

// *******************************************************************
// Name : DDRestoreSprGrdSrc.
// Desc : Restores the "Sprite Grid Source" surface and
//        reloads the BMP.
// *******************************************************************

function DDRestoreSprGrdSrc(var _sprGrdSrc: TDDSprGrdSrc): boolean;

begin
  Result := DDRestoreSprSrc(_sprGrdSrc.SprSrc);
end;

// *******************************************************************
// Name : DDFreeSprGrdSrc.
// Desc : Frees a TDDSprGrdSrc.
// *******************************************************************

procedure DDFreeSprGrdSrc(var _sprGrdSrc: TDDSprGrdSrc);

begin
  DDFreeSprSrc(_sprGrdSrc.SprSrc);
  DDFreeSurfGrid(_sprGrdSrc.Grid);
end;

// *******************************************************************
// Name : DDValidSprSrcGrd.
// Desc : Checks if a Sprite Grid Source is valid.
// *******************************************************************

function DDValidSprGrdSrc(_sprGrdSrc: TDDSprGrdSrc): boolean;

begin
  Result := (_sprGrdSrc.SprSrc.Data <> NIL) and
            (_sprGrdSrc.Grid.Count > 0);
end;

// *******************************************************************

end.
