{
 DXSprite - Sprites unit.
            Requieres DirectDraw.pas from
            DirectX 8.0 headers for Delphi.

            You can find those headers here;
            http://www.delphi-jedi.org/DelphiGraphics/
}

unit DXSprite;

// *******************************************************************

interface

uses Windows, Classes, DirectDraw, DXDraw, DXSprSrc, DXSprGrd;

type // Single sprite.
     TDDSprite = record
                   X, Y    : longint; // Read only
                   DstRect : TRect;
                   SprSrc  : pTDDSprSrc;
                 end;

     // Multiframe sprite.
     TDDMFSprite = record
                     X, Y      : longint; // Read only
                     DstRect   : TRect;
                     SrcRect   : TRect;
                     SrcWidth  : longint;
                     SrcHeight : longint;
                     CellNum   : longint;
                     SprGrdSrc : pTDDSprGrdSrc;
                   end;

// Single sprite functions.
function  DDGetSprite(_source: pTDDSprSrc): TDDSprite;
procedure DDSetSpritePos(_x, _y: longint; var _spr: TDDSprite);
procedure DDDrawSprite(_dst: IDirectDrawSurface7;
                       _spr: TDDSprite; _flags: dWord);

// Multiframe sprite functions.
function  DDGetMFSprite(_source: pTDDSprGrdSrc): TDDMFSprite;
procedure DDSetMFSpritePos(_x, _y: longint; var _spr: TDDMFSprite);
procedure DDDrawMFSprite(_dst: IDirectDrawSurface7;
                         _spr: TDDMFSprite; _flags: dWord);
procedure DDSetMFSprFrame(var _spr: TDDMFSprite; _frameno: longint);

// *******************************************************************

implementation

// *******************************************************************
// Name : DDGetSprite.
// Desc : Returns a single sprite instance.
// *******************************************************************

function DDGetSprite(_source: pTDDSprSrc): TDDSprite;

begin
  FillChar(Result, SizeOf(Result), 0);
  if _source <> NIL then
    begin
      Result.X := 0;
      Result.Y := 0;
      Result.SprSrc  := _source;
      Result.DstRect := Rect(0, 0, _source.Width - 1, _source.Height - 1);
    end;
end;

// *******************************************************************
// Name : DDSetSpritePos.
// Desc : Moves a single sprite.
// *******************************************************************

procedure DDSetSpritePos(_x, _y: longint; var _spr: TDDSprite);

begin
  if _spr.SprSrc <> NIL then
    begin
      _spr.X := _x;
      _spr.Y := _y;
      _spr.DstRect := Rect(_x, _y, _x + _spr.SprSrc^.width,
                           _y + _spr.SprSrc^.height);
    end;
end;

// *******************************************************************
// Name : DDDrawSpritePos.
// Desc : Draws a single sprite on _dst surface.
//        _dst must be DDMAINS (primary) or DDBACKS (backbuffer).
// *******************************************************************

procedure DDDrawSprite(_dst: IDirectDrawSurface7;
                       _spr: TDDSprite; _flags: dWord);

var _hr : hResult;

begin
  if _spr.SprSrc = NIL then exit;
  _hr := _dst.Blt(@_spr.DstRect, _spr.SprSrc^.Data,
                  @_spr.SprSrc^.Rect, _flags, NIL);
  if _hr = DDERR_SURFACELOST
    then DDRestoreSprSrc(_spr.SprSrc^);
end;

// *******************************************************************
// Name : DDGetMFSprite.
// Desc : Returns a multi-frame sprite instance.
// *******************************************************************

function DDGetMFSprite(_source: pTDDSprGrdSrc): TDDMFSprite;

begin
  FillChar(Result, SizeOf(Result), 0);
  if _source <> NIL then with Result do
    begin
      if _source^.Grid.Count > 0 then
        begin
          Move(_source^.Grid.Rects^, SrcRect, SizeOf(TRect));
          X := 0; Y := 0;
          SrcWidth  := SrcRect.Right  - SrcRect.Left;
          SrcHeight := SrcRect.Bottom - SrcRect.Top;
          DstRect := SrcRect;
          CellNum := 1;
        end else CellNum := 0;
      SprGrdSrc := _source;
    end;
end;

// *******************************************************************
// Name : DDSetMFSpritePos.
// Desc : Moves a multi-frame sprite.
// *******************************************************************

procedure DDSetMFSpritePos(_x, _y: longint; var _spr: TDDMFSprite);

begin
  if _spr.SprGrdSrc <> NIL then
    begin
      _spr.X := _x;
      _spr.Y := _y;
      _spr.DstRect := Rect(_x, _y, _x + _spr.SrcWidth,
                           _y + _spr.SrcHeight);
    end;
end;

// *******************************************************************
// Name : DDDrawMFSprite.
// Desc : Draws a multi-frame sprite on _dst surface.
//        _dst must be DDMAINS (primary) or DDBACKS (backbuffer).
// *******************************************************************

procedure DDDrawMFSprite(_dst: IDirectDrawSurface7;
                         _spr: TDDMFSprite; _flags: dWord);

var _hr : hResult;

begin
  if _spr.SprGrdSrc = NIL then exit;
  _hr := _dst.Blt(@_spr.DstRect, _spr.SprGrdSrc^.SprSrc.Data,
                  @_spr.SrcRect, _flags, NIL);
  if _hr = DDERR_SURFACELOST
    then DDRestoreSprSrc(_spr.SprGrdSrc^.SprSrc);
end;

// *******************************************************************
// Name : DDSetMFSprFrame.
// Desc : Changes the current frame.
//        if _frameno < 0 the next frame will be used.
// *******************************************************************

procedure DDSetMFSprFrame(var _spr: TDDMFSprite; _frameno: longint);

var _rct : ^TRect;
    _cell : longint;

begin
  with _spr do
    begin
      // Check values.
      if SprGrdSrc = NIL then exit;
      if (_frameno = 0) or (_frameno > SprGrdSrc^.Grid.Count) then exit;

      // Choose cell #.
      if _frameno < 0 then _cell := CellNum else _cell := _frameno - 1;
      if _cell >= SprGrdSrc^.Grid.Count then _cell := 0;

      // Save new frame #.
      CellNum := _cell + 1;

      // Set new frame.
      _rct := SprGrdSrc^.Grid.Rects;
      inc(longint(_rct), _cell * SizeOf(TRect));
      SrcRect := _rct^;
      SrcWidth := SrcRect.Right - SrcRect.Left;
      SrcHeight := SrcRect.Bottom - SrcRect.Top;
    end;
end;

// *******************************************************************

end.
