unit GameCore;

// ------------------------------------------------------------------

interface

uses Windows, SysUtils, Forms, Math,
     DirectInput8, DX8DI, // User input engine - DirectInput (DX8)
     DirectShow, DX8DSMP, // Music engine ------ DirectShow  (DX8)
     DirectMusic, DX8DM,  // Sound engine ------ DirectMusic (DX8)
     DirectDraw,          // 2D engine --------- DirectDraw  (DX7)
     DXDraw, DXSprite,    //
     DXSprSrc, DXSprGrd,  //
     StoryBrd;            // Storyboard

var
  GAMEPATH      : string;
  GAMEGFXPATH   : string;
  GAMESOUNDPATH : string;
  GAMEMUSICPATH : string;

  gDSMPMUSIC : TDSMP;

  GAMEREADY          : boolean = false;
  GAMEAVAILABLE      : boolean = false;
  GAMEINITIALIZED    : boolean = false;
  PLAYINGMUSIC       : boolean = false;
  USESOUNDS          : boolean = true;

  JOYSTICKAVAILABLE  : boolean = false;

const
  SOUND_REVERB    = true;
  PARAM_WINDOWED  = 'WINDOW';

  // Run stages
  RUNSTAGE_INTROINIT  = 0;
  RUNSTAGE_INTRO      = 1;
  RUNSTAGE_MENUINIT   = 2;
  RUNSTAGE_MENU       = 3;
  RUNSTAGE_OPTIONINIT = 4;
  RUNSTAGE_OPTION     = 5;
  RUNSTAGE_GAMEINIT   = 6;
  RUNSTAGE_STAGEBEGIN = 7;
  RUNSTAGE_GAME       = 8;
  RUNSTAGE_GAMEMENU   = 9;
  RUNSTAGE_INITGOVER  = 10;
  RUNSTAGE_GAMEOVER   = 11;
  RUNSTAGE_STGCINIT   = 12;
  RUNSTAGE_STGCLEAR   = 13;
  RUNSTAGE_EOGAME     = 14;

  ERR00 = 'Error initializing DirectInput.';
  ERR01 = 'Error initializing Keyboard.';
  ERR02 = 'Error initializing DirectDraw.';
  ERR03 = 'Unknown error (Not enough memory?).';
  ERR04 = 'Error initializing DirectShow.';
  ERR05 = 'Error initializing DirectMusic.'+#10+
          '(Sound will be disabled)';
  ERR06 = 'Unknown error loading file: ';
  ERR07 = #10'(File not found? Not enough memory?)';
  ERR08 = 'Unknown error.';
  ERR09 = 'Cannot set 320x200x16 screen mode!';
  ERR10 = 'Error initializing Mouse.';

  FRAME_TIME = 20;
  KEYTIME    = 250;

  TITLE_POS  = 20;
  ABOUT_POS  = 160;
  PLAYER_Y   = 165;

  KEYFRAME_000 = -700;  // Pulsar Studio Presents...
  KEYFRAME_001 = -500;  // A game by...
  KEYFRAME_002 = -300;  // Music ...
  KEYFRAME_003 = -100;  // End of titles.

var
  RUNSTAGE : longint = RUNSTAGE_INTROINIT;
  INIT_FULLSCREEN : boolean;

procedure VarInit;
procedure RenderFrame(_hWnd: hWnd);

// DirectX
procedure DXInit(_hWnd: hWnd);
procedure DXClose;

// Gfx
procedure GFXLoad(_hWnd: hWnd);
procedure GFXFree;

// Sound
procedure SoundLoad(_hWnd: hWnd);
procedure SoundFree;

// ------------------------------------------------------------------

implementation

const
  DIR_GFX    = 'GFX\';
  DIR_SOUND  = 'SOUND\';
  DIR_MUSIC  = 'MUSIC\';

  SPR_CROP_FN        = 'CROP.BMP';
  SPR_BACKGROUND_FN  = 'SPACE.BMP';
  SPR_ABOUT_FN       = 'ABOUT.BMP';
  SPR_TITLE_FN       = 'TITLE.BMP';
  SPR_INTROTEXT_FN   = 'INTROTXT.BMP';
  SPR_MENUTEXT_FN    = 'MENUTXT.BMP';
  SPR_ICON5X9_FN     = 'ICON5X9.BMP';
  SPR_PLANETS_FN     = 'PLANETS.BMP';
  SPR_STGTITLE_FN    = 'STGTITLE.BMP';
  SPR_GOPTBACK_FN    = 'GOPTBACK.BMP';
  SPR_PLAYER_FN      = 'PLAYER.BMP';
  SPR_PLAYSHOT_FN    = 'PLAYSHOT.BMP';
  SPR_NUMBERS_FN     = 'NUMBERS.BMP';
  SPR_EXPLOSION1_FN  = 'EXPLOS1.BMP';
  SPR_EXPLOSION2_FN  = 'EXPLOS2.BMP';
  SPR_EXPLOSION3_FN  = 'EXPLOS3.BMP';
  SPR_EXPLOSION4_FN  = 'EXPLOS4.BMP';
  SPR_BADGUY00_FN    = 'TIESMILE.BMP';
  SPR_BADGUY01_FN    = 'MINE.BMP';
  SPR_BADGUY02_FN    = 'THRISMIL.BMP';
  SPR_BADGUY03_FN    = 'TNKSMILE.BMP';
  SPR_BADGUY04_FN    = 'JMPSMILE.BMP';
  SPR_BADGUY05_FN    = 'BICANNON.BMP';
  SPR_BADGUY06_FN    = 'MLAUNCH.BMP';
  SPR_BADGUY07_FN    = 'PIRAMID.BMP';
  SPR_BADGUY08_FN    = 'KAMIKAZE.BMP';
  SPR_BADGUY09_FN    = 'CONTAINR.BMP';
  SPR_BGSHOOT1_FN    = 'BGSHOOT1.BMP';
  SPR_BGSHOOT2_FN    = 'BGSHOOT2.BMP';
  SPR_BGSHOOT3_FN    = 'BGSHOOT3.BMP';
  SPR_BGSHOOT4_FN    = 'BGSHOOT4.BMP';
  SPR_BGSHOOT5_FN    = 'BGSHOOT5.BMP';
  SPR_ITEMS_FN       = 'ITEMS.BMP';
  SPR_GAMEOVER_FN    = 'GAMEOVER.BMP';
  SPR_STGCLEAR_FN    = 'STGCLEAR.BMP';
  SPR_TEXT1_FN       = 'TEXT1.BMP';
  SPR_TEXT2_FN       = 'TEXT2.BMP';
  SPR_FRAME_FN       = 'FRAME.BMP';

  SND_MENUCHG_FN    = 'MENUCHG.WAV';
  SND_MENUSEL_FN    = 'MENUSEL.WAV';
  SND_STAGEINI_FN   = 'STAGEINI.WAV';
  SND_PLAYSHOT_FN   = 'PLAYSHOT.WAV';
  SND_EXPLOSION1_FN = 'EXPLOS1.WAV';
  SND_EXPLOSION2_FN = 'EXPLOS2.WAV';
  SND_EXPLOSION3_FN = 'EXPLOS3.WAV';
  SND_BGSHOOT1_FN   = 'BGSHOOT1.WAV';
  SND_BGSHOOT2_FN   = 'BGSHOOT2.WAV';
  SND_ITEM_FN       = 'ITEM.WAV';
  SND_MLAUNCH_FN    = 'MLAUNCH.WAV';
  SND_BIGLASER_FN   = 'BIGLASER.WAV';

  MUSIC_INTRO = 'EVANGLN.MID';
  MUSIC_GAME  = 'SALAMAND.MID';
  MUSIC_END   = 'NEMESIS3.MID';

  MUSICLOOPTESTIN     = 60;
  MAXVOLUME           = 25;
  DEFPLAYERSHOOTS     = 2;
  MAXPLAYERSHOOTS     = 4;
  MAXBADGUYS          = 150;
  RESURRECTIONTIME    = 125;
  INVULNERABILITYTIME = 125;

  TIEFIGHTER_RELOADTIME = 100;
  THRI_SMILE_RELOADTIME = 60;
  TNK_SMILE_RELOADTIME  = 60;
  JUMP_SMILE_RELOADTIME = 50;
  BI_CANNON_RELOADTIME  = 30;
  MLAUNCHER_RELOADTIME  = 42;
  PIRAMID_RELOADTIME    = 60;
  PIRAMID_RELOADTIME2   = 40;
  PIRAMID_FIRETIME      = 60;
  YLIMIT_FOR_S00TING    = 100;

  KEYTIMEOUT = 250;
  JOYTIMEOUT = 250;

type
  TBADGUY = record
              Spr    : TDDMFSprite;
              Clan   : longint; // Type of bad-guy.
              Time   : longint; // Time of life.
              Life   : longint; // Life of bad-guy.
              OX, OY : longint; // Org X - Org Y.
              Alive  : boolean;
              SBPar1 : longint;
              Param1 : longint;
              Param2 : longint;
            end;

var
  SPR_CROP_SRC : TDDSprSrc;
  SPR_CROP_000 : TDDSprite;
  SPR_CROP_001 : TDDSprite;
  SPR_CROP_ANI : longint;

  SPR_BACKGROUND_SRC : TDDSprSrc;
  SPR_BACKGROUND_000 : TDDSprite;
  SPR_BACKGROUND_001 : TDDSprite;
  SPR_BACKGROUND_ANI : longint;

  SPR_ABOUT_SRC : TDDSprSrc;
  SPR_ABOUT_000 : TDDSprite;
  SPR_ABOUT_ANI : longint;

  SPR_INTROTEXT_SRC : TDDSprGrdSrc;
  SPR_INTROTEXT_000 : TDDMFSprite;

  SPR_MENUTEXT_SRC : TDDSprGrdSrc;
  SPR_MENUTEXT_000 : TDDMFSprite;
  SPR_MENUTEXT_001 : TDDMFSprite;
  SPR_MENUTEXT_002 : TDDMFSprite;
  SPR_MENUTEXT_003 : TDDMFSprite;
  SPR_MENUTEXT_004 : TDDMFSprite;
  SPR_MENUTEXT_005 : TDDMFSprite;

  SPR_ICON5X9_SRC : TDDSprGrdSrc;
  SPR_ICON5X9_000 : TDDMFSprite;
  SPR_ICON5X9_001 : TDDMFSprite;

  SPR_TITLE_SRC : TDDSprSrc;
  SPR_TITLE_000 : TDDSprite;
  SPR_TITLE_ANI : longint;

  SPR_PLANETS_SRC : TDDSprGrdSrc;
  SPR_PLANETS_000 : TDDMFSprite;
  SPR_PLANETS_ANI : longint;

  SPR_STGTITLE_SRC : TDDSprSrc;
  SPR_STGTITLE_000 : TDDSprite;

  SPR_GOPTBACK_SRC : TDDSprSrc;
  SPR_GOPTBACK_000 : TDDSprite;

  SPR_PLAYER_SRC : TDDSprGrdSrc;
  SPR_PLAYER_000 : TDDMFSprite;
  SPR_PLAYER_ANI : longint = 0;

  SPR_PLAYSHOT_SRC : TDDSprGrdSrc;
  SPR_PLAYSHOT_00X : array[0..MAXPLAYERSHOOTS-1] of record
                                                      SPR   : TDDMFSprite;
                                                      InUse : boolean;
                                                    end;

  SPR_NUMBERS_SRC : TDDSprGrdSrc;
  SPR_NUMBERS_000 : TDDMFSprite;

  SPR_EXPLOSION1_SRC : TDDSprGrdSrc;
  SPR_EXPLOSION2_SRC : TDDSprGrdSrc;
  SPR_EXPLOSION3_SRC : TDDSprGrdSrc;
  SPR_EXPLOSION4_SRC : TDDSprGrdSrc;
  SPR_BADGUY00_SRC   : TDDSprGrdSrc;
  SPR_BADGUY01_SRC   : TDDSprGrdSrc;
  SPR_BADGUY02_SRC   : TDDSprGrdSrc;
  SPR_BADGUY03_SRC   : TDDSprGrdSrc;
  SPR_BADGUY04_SRC   : TDDSprGrdSrc;
  SPR_BADGUY05_SRC   : TDDSprGrdSrc;
  SPR_BADGUY06_SRC   : TDDSprGrdSrc;
  SPR_BADGUY07_SRC   : TDDSprGrdSrc;
  SPR_BADGUY08_SRC   : TDDSprGrdSrc;
  SPR_BADGUY09_SRC   : TDDSprGrdSrc;
  SPR_BGSHOOT1_SRC   : TDDSprGrdSrc;
  SPR_BGSHOOT2_SRC   : TDDSprGrdSrc;
  SPR_BGSHOOT3_SRC   : TDDSprGrdSrc;
  SPR_BGSHOOT4_SRC   : TDDSprGrdSrc;
  SPR_BGSHOOT5_SRC   : TDDSprGrdSrc;
  SPR_ITEMS_SRC      : TDDSprGrdSrc;

  BADGUYS : record
              BG    : array[0..MAXBADGUYS-1] of TBADGUY;
              Count : longint; // Badguys on screen.
            end;

  SPR_GAMEOVER_SRC : TDDSprSrc;
  SPR_GAMEOVER_000 : TDDSprite;
  SPR_GAMEOVER_ANI : longint;

  SPR_STGCLEAR_SRC : TDDSprSrc;
  SPR_STGCLEAR_000 : TDDSprite;
  SPR_STGCLEAR_ANI : longint;

  SPR_TEXT1_SRC    : TDDSprSrc;
  SPR_TEXT1_000    : TDDSprite;

  SPR_TEXT2_SRC    : TDDSprSrc;
  SPR_TEXT2_000    : TDDSprite;

  SPR_FRAME_SRC    : TDDSprSrc;
  SPR_FRAME_000    : TDDSprite;

  SND_MENUCHG_SEG    : IDirectMusicSegment8 = NIL;
  SND_MENUSEL_SEG    : IDirectMusicSegment8 = NIL;
  SND_STAGEINI_SEG   : IDirectMusicSegment8 = NIL;
  SND_PLAYSHOT_SEG   : IDirectMusicSegment8 = NIL;
  SND_EXPLOSION1_SEG : IDirectMusicSegment8 = NIL;
  SND_EXPLOSION2_SEG : IDirectMusicSegment8 = NIL;
  SND_EXPLOSION3_SEG : IDirectMusicSegment8 = NIL;
  SND_BGSHOOT1_SEG   : IDirectMusicSegment8 = NIL;
  SND_BGSHOOT2_SEG   : IDirectMusicSegment8 = NIL;
  SND_ITEM_SEG       : IDirectMusicSegment8 = NIL;
  SND_MLAUNCH_SEG    : IDirectMusicSegment8 = NIL;
  SND_BIGLASER_SEG   : IDirectMusicSegment8 = NIL;

  FRAMECONTROL     : dWord;
  MENUSELECTION    : longint = 0;
  OPTIONSELECTION  : longint = 0;
  GAMEOPTSELECTION : longint = 0;
  MUSICLOOPTESTCNT : longint = 0;
  STAGENO          : longint = 0;
  PLAYER_X         : longint = 0;
  STAGETIME        : longint = 0;
  PLAYERSCORE      : longint = 0;
  PLAYERLIFES      : longint = 0;
  PLAYERKILLED     : boolean = false;
  FASTSHOOT        : boolean = false;
  INVULNERABILITY  : longint = 0;
  NUMPLAYERSHOOTS  : longint;
  BLUEITEMCOUNT    : longint;

  STORYBOARDCUR    : pointer = NIL;

  MUSICVOLUME      : longint = 20;
  SOUNDVOLUME      : longint = MAXVOLUME;

  // Keyboard Timed Inputs
  KTI_UP    : TTimedInput;
  KTI_DOWN  : TTimedInput;
  KTI_LEFT  : TTimedInput;
  KTI_RIGHT : TTimedInput;
  KTI_SHOT  : TTimedInput;
  KTI_ESC   : TTimedInput;
  KTI_SF11  : TTimedInput;
  KTI_SF12  : TTimedInput;

  // Joystick Timed Inputs
  JTI_UP    : TTimedInput;
  JTI_DOWN  : TTimedInput;
  JTI_LEFT  : TTimedInput;
  JTI_RIGHT : TTimedInput;
  JTI_SHOT  : TTimedInput;
  JTI_ESC   : TTimedInput;

// ------------------------------------------------------------------

procedure ErrHalt(_hWnd: hWnd; _msg: string);

begin
  MessageBox(_hWnd, pChar(_msg), 'Error', MB_ICONERROR);
  Application.Terminate;
end;

// ------------------------------------------------------------------

procedure ErrWarning(_hWnd: hWnd; _msg: string);

begin
  MessageBox(_hWnd, pChar(_msg), 'Error/Warning', MB_ICONWARNING);
end;

// ------------------------------------------------------------------

procedure ScaleRect(var _rect: TRect; _perc: integer);

var _x, _y : longint;

begin
  _x := _rect.Right  - _rect.Left + 1;
  _y := _rect.Bottom - _rect.Top  + 1;
  _x := (_perc * _x) div 100;
  _y := (_perc * _y) div 100;
  _rect.Left   := _rect.Left   - _x;
  _rect.Right  := _rect.Right  + _x;
  _rect.Top    := _rect.Top    - _y;
  _rect.Bottom := _rect.Bottom + _y;
end;

// ------------------------------------------------------------------

function RectCollision(var _r1, _r2: TRect): boolean;

begin
  Result := not ((_r1.Right  < _r2.Left) or (_r1.Bottom < _r2.Top) or
                 (_r2.Right  < _r1.Left) or (_r2.Bottom < _r1.Top));
end;

// ------------------------------------------------------------------

procedure MinMax(var _x: longint; _inc, _min, _max: longint;
                     _playSnd: boolean);

begin
  inc(_x, _inc);
  if _x > _max then _x := _min;
  if _x < _min then _x := _max;
  if _playSnd then DMPlay(SND_MENUCHG_SEG);
end;

// ------------------------------------------------------------------

procedure ResetTimedInputs;

begin
  // Keyboard.
  TimedInputReset(KTI_UP);
  TimedInputReset(KTI_DOWN);
  TimedInputReset(KTI_LEFT);
  TimedInputReset(KTI_RIGHT);
  TimedInputReset(KTI_SHOT);
  TimedInputReset(KTI_ESC);
  TimedInputReset(KTI_SF11);
  TimedInputReset(KTI_SF12);

  // Joystick.
  TimedInputReset(JTI_UP);
  TimedInputReset(JTI_DOWN);
  TimedInputReset(JTI_LEFT);
  TimedInputReset(JTI_RIGHT);
  TimedInputReset(JTI_SHOT);
  TimedInputReset(JTI_ESC);
end;

// ------------------------------------------------------------------

procedure VarInit;

begin
  // Get application path.
  GAMEPATH      := ExtractFilePath(paramstr(0));
  GAMEGFXPATH   := GAMEPATH + DIR_GFX;
  GAMESOUNDPATH := GAMEPATH + DIR_SOUND;
  GAMEMUSICPATH := GAMEPATH + DIR_MUSIC;
  FRAMECONTROL  := GetTickCount;

  // Reset DSMP.
  dsmpReset(gDSMPMUSIC);

  // Clear sprite sources.
  DDResetSprSrc(SPR_BACKGROUND_SRC);
  DDResetSprSrc(SPR_TITLE_SRC);
  DDResetSprSrc(SPR_ABOUT_SRC);
  DDResetSprGrdSrc(SPR_INTROTEXT_SRC);
end;

// ------------------------------------------------------------------

procedure DXInit(_hWnd: hWnd);

begin
  // Initialize DirectMusic.
  if not DMInit(_hWnd, SOUND_REVERB) then
    begin
      ErrWarning(_hWnd, ERR05);
      USESOUNDS := false;
    end;

  // Initialize DirectInput.
  ResetTimedInputs;  
  if not DIInit then
    begin ErrHalt(_hWnd, ERR00); exit; end;
  if not DIInitKeyboard(_hWnd) then
    begin ErrHalt(_hWnd, ERR01); exit; end;
  if not DIInitMouse(_hWnd) then
    begin ErrHalt(_hWnd, ERR10); exit; end;
  JOYSTICKAVAILABLE := DIInitJoystick(_hWnd);

  // Initialize DirectDraw.
  if not DDInit(NIL, _hWnd, INIT_FULLSCREEN) then
    begin ErrHalt(_hWnd, ERR02); exit; end;

  if INIT_FULLSCREEN then
    begin
      if not DDSetDisplayMode(320, 200, 16, 0) then
        begin ErrHalt(_hWnd, ERR09); exit; end;
      if not DDCreateFullScrBuffers then
        begin ErrHalt(_hWnd, ERR03); exit; end;
    end else if not DDCreateWindowBuffers(_hWnd) then
               begin ErrHalt(_hWnd, ERR03); exit; end;

  // Set clipper.
  if not DDSetClipper(_hWnd, DDBACKS) then
    begin ErrHalt(_hWnd, ERR08); exit; end;
end;

// ------------------------------------------------------------------

procedure DXClose;

begin
  dsmpFree(gDSMPMUSIC); // Free directShow.
  DDFreeMainBuffers; // Free DirectDraw main buffers.
  DDClose; // Close DirectDraw.
  DIClose; // Close DirectInput.
  DMClose; // Close DirectMusic.
end;

// ------------------------------------------------------------------

procedure SprGrdLoad(_hWnd: hWnd; _fn: string;
                     var _spr: TDDSprGrdSrc; _ck: cardinal);

begin
  _spr := DDCreateSprGrdSrc(GAMEGFXPATH + _fn);
  if not DDValidSprGrdSrc(_spr)
    then ErrHalt(_hWnd, ERR06 + _fn + ERR07);
  if _ck <> CLR_INVALID then
    if not DDSetColorKey(_spr.SprSrc.Data, _ck)
      then ErrHalt(_hWnd, ERR08);
end;

// ------------------------------------------------------------------

procedure SprLoad(_hWnd: hWnd; _fn: string;
                  var _spr: TDDSprSrc; _ck: dWord);

begin
  _spr := DDCreateSprSrc(GAMEGFXPATH + _fn);
  if not DDValidSprSrc(_spr)
    then ErrHalt(_hWnd, ERR06 + _fn + ERR07);
  if _ck <> CLR_INVALID then
    if not DDSetColorKey(_spr.Data, _ck)
      then ErrHalt(_hWnd, ERR08);
end;

// ------------------------------------------------------------------

procedure GFXLoad(_hWnd: hWnd);

var _cnt : longint;

begin
  // Load Sprite Sources
  SprLoad(_hWnd, SPR_CROP_FN,       SPR_CROP_SRC,       CLR_INVALID);
  SprLoad(_hWnd, SPR_BACKGROUND_FN, SPR_BACKGROUND_SRC, CLR_INVALID);
  SprLoad(_hWnd, SPR_ABOUT_FN,      SPR_ABOUT_SRC,      $0000FF);
  SprLoad(_hWnd, SPR_TITLE_FN,      SPR_TITLE_SRC,      $FF0000);
  SprLoad(_hWnd, SPR_STGTITLE_FN,   SPR_STGTITLE_SRC,   $0000FF);
  SprLoad(_hWnd, SPR_GOPTBACK_FN,   SPR_GOPTBACK_SRC,   CLR_INVALID);
  SprLoad(_hWnd, SPR_GAMEOVER_FN,   SPR_GAMEOVER_SRC,   CLR_INVALID);
  SprLoad(_hWnd, SPR_STGCLEAR_FN,   SPR_STGCLEAR_SRC,   CLR_INVALID);
  SprLoad(_hWnd, SPR_TEXT1_FN,      SPR_TEXT1_SRC,      $0000FF);
  SprLoad(_hWnd, SPR_TEXT2_FN,      SPR_TEXT2_SRC,      $0000FF);
  SprLoad(_hWnd, SPR_FRAME_FN,      SPR_FRAME_SRC,      $FFFFFF);

  // Load Sprite-Grid Sources.
  SprGrdLoad(_hWnd, SPR_MENUTEXT_FN,   SPR_MENUTEXT_SRC,   $0000FF);
  SprGrdLoad(_hWnd, SPR_INTROTEXT_FN,  SPR_INTROTEXT_SRC,  $0000FF);
  SprGrdLoad(_hWnd, SPR_ICON5X9_FN,    SPR_ICON5X9_SRC,    $00FFFF);
  SprGrdLoad(_hWnd, SPR_PLANETS_FN,    SPR_PLANETS_SRC,    CLR_INVALID);
  SprGrdLoad(_hWnd, SPR_PLAYER_FN,     SPR_PLAYER_SRC,     $00FF00);
  SprGrdLoad(_hWnd, SPR_PLAYSHOT_FN,   SPR_PLAYSHOT_SRC,   $0000FF);
  SprGrdLoad(_hWnd, SPR_NUMBERS_FN,    SPR_NUMBERS_SRC,    $0000FF);
  SprGrdLoad(_hWnd, SPR_EXPLOSION1_FN, SPR_EXPLOSION1_SRC, $00FF00);
  SprGrdLoad(_hWnd, SPR_EXPLOSION2_FN, SPR_EXPLOSION2_SRC, $00FF00);
  SprGrdLoad(_hWnd, SPR_EXPLOSION3_FN, SPR_EXPLOSION3_SRC, $00FF00);
  SprGrdLoad(_hWnd, SPR_EXPLOSION4_FN, SPR_EXPLOSION4_SRC, $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY00_FN,   SPR_BADGUY00_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY01_FN,   SPR_BADGUY01_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY02_FN,   SPR_BADGUY02_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY03_FN,   SPR_BADGUY03_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY04_FN,   SPR_BADGUY04_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY05_FN,   SPR_BADGUY05_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY06_FN,   SPR_BADGUY06_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY07_FN,   SPR_BADGUY07_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY08_FN,   SPR_BADGUY08_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BADGUY09_FN,   SPR_BADGUY09_SRC,   $00FFFF);
  SprGrdLoad(_hWnd, SPR_BGSHOOT1_FN,   SPR_BGSHOOT1_SRC,   $0000FF);
  SprGrdLoad(_hWnd, SPR_BGSHOOT2_FN,   SPR_BGSHOOT2_SRC,   $0000FF);
  SprGrdLoad(_hWnd, SPR_BGSHOOT3_FN,   SPR_BGSHOOT3_SRC,   $FF0000);
  SprGrdLoad(_hWnd, SPR_BGSHOOT4_FN,   SPR_BGSHOOT4_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_BGSHOOT5_FN,   SPR_BGSHOOT5_SRC,   $00FF00);
  SprGrdLoad(_hWnd, SPR_ITEMS_FN,      SPR_ITEMS_SRC,      $00FFFF);

  // Initialize Sprites
  FillChar(BADGUYS, SizeOf(BADGUYS), 0);
  SPR_CROP_000       := DDGetSprite(@SPR_CROP_SRC);
  SPR_CROP_001       := DDGetSprite(@SPR_CROP_SRC);
  SPR_BACKGROUND_000 := DDGetSprite(@SPR_BACKGROUND_SRC);
  SPR_BACKGROUND_001 := DDGetSprite(@SPR_BACKGROUND_SRC);
  SPR_ABOUT_000      := DDGetSprite(@SPR_ABOUT_SRC);
  SPR_TITLE_000      := DDGetSprite(@SPR_TITLE_SRC);
  SPR_INTROTEXT_000  := DDGetMFSprite(@SPR_INTROTEXT_SRC);
  SPR_MENUTEXT_000   := DDGetMFSprite(@SPR_MENUTEXT_SRC);
  SPR_MENUTEXT_001   := DDGetMFSprite(@SPR_MENUTEXT_SRC);
  SPR_MENUTEXT_002   := DDGetMFSprite(@SPR_MENUTEXT_SRC);
  SPR_MENUTEXT_003   := DDGetMFSprite(@SPR_MENUTEXT_SRC);
  SPR_MENUTEXT_004   := DDGetMFSprite(@SPR_MENUTEXT_SRC);
  SPR_MENUTEXT_005   := DDGetMFSprite(@SPR_MENUTEXT_SRC);
  SPR_ICON5X9_000    := DDGetMFSprite(@SPR_ICON5X9_SRC);
  SPR_ICON5X9_001    := DDGetMFSprite(@SPR_ICON5X9_SRC);
  SPR_PLANETS_000    := DDGetMFSprite(@SPR_PLANETS_SRC);
  SPR_STGTITLE_000   := DDGetSprite(@SPR_STGTITLE_SRC);
  SPR_GOPTBACK_000   := DDGetSprite(@SPR_GOPTBACK_SRC);
  SPR_PLAYER_000     := DDGetMFSprite(@SPR_PLAYER_SRC);
  SPR_NUMBERS_000    := DDGetMFSprite(@SPR_NUMBERS_SRC);
  SPR_GAMEOVER_000   := DDGetSprite(@SPR_GAMEOVER_SRC);
  SPR_STGCLEAR_000   := DDGetSprite(@SPR_STGCLEAR_SRC);
  SPR_TEXT1_000      := DDGetSprite(@SPR_TEXT1_SRC);
  SPR_TEXT2_000      := DDGetSprite(@SPR_TEXT2_SRC);
  SPR_FRAME_000      := DDGetSprite(@SPR_FRAME_SRC);
  for _cnt := 0 to MAXPLAYERSHOOTS - 1 do
    begin
      SPR_PLAYSHOT_00X[_cnt].SPR := DDGetMFSprite(@SPR_PLAYSHOT_SRC);
      SPR_PLAYSHOT_00X[_cnt].InUse := false;
      DDSetMFSprFrame(SPR_PLAYSHOT_00X[_cnt].SPR, 2);
    end;
end;

// ------------------------------------------------------------------

procedure GFXFree;

begin
  // Free Sprite Sources.
  DDFreeSprSrc(SPR_CROP_SRC);
  DDFreeSprSrc(SPR_STGCLEAR_SRC);
  DDFreeSprSrc(SPR_GAMEOVER_SRC);
  DDFreeSprGrdSrc(SPR_ITEMS_SRC);  
  DDFreeSprGrdSrc(SPR_BGSHOOT5_SRC);
  DDFreeSprGrdSrc(SPR_BGSHOOT4_SRC);
  DDFreeSprGrdSrc(SPR_BGSHOOT3_SRC);
  DDFreeSprGrdSrc(SPR_BGSHOOT2_SRC);
  DDFreeSprGrdSrc(SPR_BGSHOOT1_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY09_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY08_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY07_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY06_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY05_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY04_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY03_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY02_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY01_SRC);
  DDFreeSprGrdSrc(SPR_BADGUY00_SRC);
  DDFreeSprGrdSrc(SPR_EXPLOSION4_SRC);
  DDFreeSprGrdSrc(SPR_EXPLOSION3_SRC);
  DDFreeSprGrdSrc(SPR_EXPLOSION2_SRC);
  DDFreeSprGrdSrc(SPR_EXPLOSION1_SRC);
  DDFreeSprGrdSrc(SPR_NUMBERS_SRC);
  DDFreeSprGrdSrc(SPR_PLAYSHOT_SRC);
  DDFreeSprGrdSrc(SPR_PLAYER_SRC);
  DDFreeSprSrc(SPR_GOPTBACK_SRC);
  DDFreeSprSrc(SPR_STGTITLE_SRC);
  DDFreeSprGrdSrc(SPR_PLANETS_SRC);
  DDFreeSprGrdSrc(SPR_ICON5X9_SRC);
  DDFreeSprGrdSrc(SPR_MENUTEXT_SRC);
  DDFreeSprGrdSrc(SPR_INTROTEXT_SRC);
  DDFreeSprSrc(SPR_TITLE_SRC);
  DDFreeSprSrc(SPR_ABOUT_SRC);
  DDFreeSprSrc(SPR_BACKGROUND_SRC);
  DDFreeSprSrc(SPR_TEXT1_SRC);
  DDFreeSprSrc(SPR_TEXT2_SRC);
  DDFreeSprSrc(SPR_FRAME_SRC);
end;

// ------------------------------------------------------------------

procedure UpdateAudioVolume;

var _m, _v : longint;

begin
  _m := (-MAXVOLUME+MUSICVOLUME);
  _v := (-MAXVOLUME+SOUNDVOLUME);
  if _m = -MAXVOLUME
    then _m := DSMINAUDIOVOLUME else _m := _m * (3000 div MAXVOLUME);
  if _v = -MAXVOLUME
    then _v := DMMINAUDIOVOLUME else _v := _v * (3000 div MAXVOLUME);
  if PLAYINGMUSIC
    then gDSMPMUSIC.BasicAudio.Put_Volume(_m); // Music.
  if USESOUNDS
    then DM8AudioPath.SetVolume(_v, 0);  // Sound.
end;

// ------------------------------------------------------------------

procedure SndLoad(_hWnd: hWnd;
                  _fn: string; var _seg: IDirectMusicSegment8);

begin
  _seg := DMLoadWavFile(GAMESOUNDPATH + _fn);
  if _seg = NIL then ErrHalt(_hWnd, ERR06 + _fn + ERR07);
end;

// ------------------------------------------------------------------

procedure SoundLoad(_hWnd: hWnd);

begin
  if not USESOUNDS then exit;

  SndLoad(_hWnd, SND_MENUCHG_FN, SND_MENUCHG_SEG);
  SndLoad(_hWnd, SND_MENUSEL_FN, SND_MENUSEL_SEG);
  SndLoad(_hWnd, SND_STAGEINI_FN, SND_STAGEINI_SEG);
  SndLoad(_hWnd, SND_PLAYSHOT_FN, SND_PLAYSHOT_SEG);
  SndLoad(_hWnd, SND_EXPLOSION1_FN, SND_EXPLOSION1_SEG);
  SndLoad(_hWnd, SND_EXPLOSION2_FN, SND_EXPLOSION2_SEG);
  SndLoad(_hWnd, SND_EXPLOSION3_FN, SND_EXPLOSION3_SEG);
  SndLoad(_hWnd, SND_BGSHOOT1_FN, SND_BGSHOOT1_SEG);
  SndLoad(_hWnd, SND_BGSHOOT2_FN, SND_BGSHOOT2_SEG);
  SndLoad(_hWnd, SND_ITEM_FN, SND_ITEM_SEG);
  SndLoad(_hWnd, SND_MLAUNCH_FN, SND_MLAUNCH_SEG);
  SndLoad(_hWnd, SND_BIGLASER_FN, SND_BIGLASER_SEG);
end;

// ------------------------------------------------------------------

procedure SoundFree;

begin
  DMFreeSegment(SND_MENUCHG_SEG);
  DMFreeSegment(SND_MENUSEL_SEG);
  DMFreeSegment(SND_STAGEINI_SEG);
  DMFreeSegment(SND_PLAYSHOT_SEG);
  DMFreeSegment(SND_EXPLOSION1_SEG);
  DMFreeSegment(SND_EXPLOSION2_SEG);
  DMFreeSegment(SND_EXPLOSION3_SEG);  
  DMFreeSegment(SND_BGSHOOT1_SEG);
  DMFreeSegment(SND_BGSHOOT2_SEG);
  DMFreeSegment(SND_ITEM_SEG);
  DMFreeSegment(SND_MLAUNCH_SEG);
  DMFreeSegment(SND_BIGLASER_SEG);
end;

// ------------------------------------------------------------------

procedure MusicLoopTest;

var _current, _stop : int64;

begin
  if (not GAMEINITIALIZED) or (not PLAYINGMUSIC) or (not GAMEAVAILABLE)
    then exit;

  gDSMPMUSIC.MediaSeeking.GetPositions(_current, _stop);
  if _current = _stop then
    begin
      _current := 0;
      gDSMPMUSIC.MediaSeeking.SetPositions(
        _current, AM_SEEKING_AbsolutePositioning,
        _current, AM_SEEKING_NoPositioning);
    end;
end;

// ------------------------------------------------------------------

procedure InitializeIntro;

begin
  // Initialize sprites.
  SPR_CROP_ANI := 0;
  SPR_BACKGROUND_ANI := 0;
  SPR_ABOUT_ANI := 1300;
  SPR_TITLE_ANI := -1000;
  DDSetSpritePos(0, 0, SPR_TEXT1_000);
  DDSetSpritePos(0, 0, SPR_TEXT2_000);
  DDSetSpritePos(0, 200 - SPR_FRAME_SRC.Height + 1, SPR_FRAME_000);
  DDSetSpritePos(0, 0, SPR_CROP_000);
  DDSetSpritePos(0, 100, SPR_CROP_001);
  DDSetSpritePos(0, SPR_BACKGROUND_ANI, SPR_BACKGROUND_000);
  DDSetSpritePos(0, SPR_BACKGROUND_ANI + SPR_BACKGROUND_SRC.Height, SPR_BACKGROUND_001);
  DDSetSpritePos(0, SPR_ABOUT_ANI, SPR_ABOUT_000);
  DDSetSpritePos(0, SPR_TITLE_ANI, SPR_TITLE_000);
  DDSetMFSpritePos(160 - (SPR_INTROTEXT_SRC.SprSrc.Width shr 1),
                   70, SPR_INTROTEXT_000);
  DDSetMFSprFrame(SPR_INTROTEXT_000, 1);

  // Play Music.
  if dsmpLoadFile(gDSMPMUSIC, GAMEMUSICPATH + MUSIC_INTRO) then
    begin
      UpdateAudioVolume;
      gDSMPMUSIC.MediaControl.Run;
      PLAYINGMUSIC := true;
    end;

  // Update Audio volume.
  UpdateAudioVolume;
end;

// ------------------------------------------------------------------

procedure InitializeMainMenu;

var _x : longint;

begin
  MENUSELECTION := 0;
  OPTIONSELECTION := 0;
  SPR_ABOUT_ANI := ABOUT_POS;
  SPR_TITLE_ANI := TITLE_POS;
  DDSetSpritePos(0, SPR_ABOUT_ANI, SPR_ABOUT_000);
  DDSetSpritePos(0, SPR_TITLE_ANI, SPR_TITLE_000);
  _x := 160 - (SPR_MENUTEXT_SRC.SprSrc.Width shr 1);
  DDSetMFSpritePos(_x,  90, SPR_MENUTEXT_000);
  DDSetMFSpritePos(_x, 105, SPR_MENUTEXT_001);
  DDSetMFSpritePos(_x, 120, SPR_MENUTEXT_002);
  DDSetMFSpritePos(_x - 60,  90, SPR_MENUTEXT_003);
  DDSetMFSpritePos(_x - 60, 105, SPR_MENUTEXT_004);
  DDSetMFSpritePos(_x, 120, SPR_MENUTEXT_005);
end;

// ------------------------------------------------------------------

procedure UpdateMainMenu;

begin
  DDSetMFSprFrame(SPR_MENUTEXT_000, 1);
  DDSetMFSprFrame(SPR_MENUTEXT_001, 3);
  DDSetMFSprFrame(SPR_MENUTEXT_002, 5);
  Case MENUSELECTION of
    0 : DDSetMFSprFrame(SPR_MENUTEXT_000, -1);
    1 : DDSetMFSprFrame(SPR_MENUTEXT_001, -1);
    2 : DDSetMFSprFrame(SPR_MENUTEXT_002, -1);
  end;
end;

// ------------------------------------------------------------------

procedure UpdateOptions;

begin
  DDSetMFSprFrame(SPR_MENUTEXT_003, 7);
  DDSetMFSprFrame(SPR_MENUTEXT_004, 9);
  DDSetMFSprFrame(SPR_MENUTEXT_005, 11);
  Case OPTIONSELECTION of
    0 : DDSetMFSprFrame(SPR_MENUTEXT_003, -1);
    1 : DDSetMFSprFrame(SPR_MENUTEXT_004, -1);
    2 : DDSetMFSprFrame(SPR_MENUTEXT_005, -1);
  end;
end;

// ------------------------------------------------------------------

procedure UpdateGameOptions;

begin
  DDSetMFSprFrame(SPR_MENUTEXT_002, 13);
  DDSetMFSprFrame(SPR_MENUTEXT_003, 7);
  DDSetMFSprFrame(SPR_MENUTEXT_004, 9);
  DDSetMFSprFrame(SPR_MENUTEXT_005, 11);
  Case GAMEOPTSELECTION of
    0 : DDSetMFSprFrame(SPR_MENUTEXT_003, -1);
    1 : DDSetMFSprFrame(SPR_MENUTEXT_004, -1);
    2 : DDSetMFSprFrame(SPR_MENUTEXT_005, -1);
    3 : DDSetMFSprFrame(SPR_MENUTEXT_002, -1);
  end;
end;

// ------------------------------------------------------------------

procedure InitializeStage;

begin
  // Load and play music.
  dsmpLoadFile(gDSMPMUSIC, GAMEMUSICPATH + MUSIC_GAME);

  // Initialize sprites.
  SPR_BACKGROUND_ANI := 0;
  SPR_PLANETS_ANI := 0;
  DDSetSpritePos(0, 0, SPR_STGTITLE_000);
  DDSetSpritePos(0, 0, SPR_GOPTBACK_000);
  DDSetMFSpritePos(160 - (SPR_MENUTEXT_SRC.SprSrc.Width shr 1),
                   135, SPR_MENUTEXT_002);
  DDSetMFSpritePos(1, 100 - (SPR_PLANETS_000.SrcHeight shr 1),
                   SPR_PLANETS_000);
  DDSetMFSprFrame(SPR_PLAYER_000, 1);
  PLAYER_X := 160 - (SPR_PLAYER_000.SrcWidth shr 1);
  DDSetMFSpritePos(PLAYER_X, PLAYER_Y, SPR_PLAYER_000);
  DDSetMFSprFrame(SPR_PLANETS_000, STAGENO);
end;

// ------------------------------------------------------------------

procedure RenderOptions;

var _cnt : longint;

begin
  DDSetMFSprFrame(SPR_ICON5X9_000, 1);
  DDSetMFSprFrame(SPR_ICON5X9_001, 1);
  for _cnt := 1 to MAXVOLUME do
    begin
      if _cnt = MUSICVOLUME + 1 then DDSetMFSprFrame(SPR_ICON5X9_000, -1);
      if _cnt = SOUNDVOLUME + 1 then DDSetMFSprFrame(SPR_ICON5X9_001, -1);
      DDSetMFSpritePos(150 + (_cnt * 5),  90, SPR_ICON5X9_000);
      DDDrawMFSprite(DDBACKS, SPR_ICON5X9_000, DDBLT_KEYSRC);
      DDSetMFSpritePos(150 + (_cnt * 5), 105, SPR_ICON5X9_001);
      DDDrawMFSprite(DDBACKS, SPR_ICON5X9_001, DDBLT_KEYSRC);
    end;
  DDDrawMFSprite(DDBACKS, SPR_MENUTEXT_003, DDBLT_KEYSRC);
  DDDrawMFSprite(DDBACKS, SPR_MENUTEXT_004, DDBLT_KEYSRC);
  DDDrawMFSprite(DDBACKS, SPR_MENUTEXT_005, DDBLT_KEYSRC);
end;

// ------------------------------------------------------------------

procedure RenderIntro;

begin
  if RUNSTAGE = RUNSTAGE_INTRO then
    begin
      if (SPR_CROP_ANI < 100) then inc(SPR_CROP_ANI);
      if (SPR_TITLE_ANI > KEYFRAME_003) and (SPR_CROP_ANI < 200)
        then inc(SPR_CROP_ANI);
      if SPR_ABOUT_ANI > ABOUT_POS
        then dec(SPR_ABOUT_ANI)
          else RUNSTAGE := RUNSTAGE_MENUINIT;
      if SPR_TITLE_ANI < TITLE_POS then inc(SPR_TITLE_ANI);
    end;

  dec(SPR_BACKGROUND_ANI);
  if (SPR_BACKGROUND_ANI div 3) + SPR_BACKGROUND_SRC.Height = 0
    then SPR_BACKGROUND_ANI := 0;
  DDSetSpritePos(0, 0 - (SPR_CROP_ANI shr 1), SPR_CROP_000);
  DDSetSpritePos(0, 100 + (SPR_CROP_ANI shr 1), SPR_CROP_001);
  DDSetSpritePos(0, (SPR_BACKGROUND_ANI div 3), SPR_BACKGROUND_000);
  DDSetSpritePos(0, (SPR_BACKGROUND_ANI div 3) +
                     SPR_BACKGROUND_SRC.Height, SPR_BACKGROUND_001);
  DDSetSpritePos(0, SPR_ABOUT_ANI, SPR_ABOUT_000);
  DDSetSpritePos(0, SPR_TITLE_ANI, SPR_TITLE_000);
  DDDrawSprite(DDBACKS, SPR_BACKGROUND_000, 0);
  DDDrawSprite(DDBACKS, SPR_BACKGROUND_001, 0);
  DDDrawSprite(DDBACKS, SPR_ABOUT_000, DDBLT_KEYSRC);
  DDDrawSprite(DDBACKS, SPR_TITLE_000, DDBLT_KEYSRC);

  if (SPR_TITLE_ANI > KEYFRAME_000) and (SPR_TITLE_ANI < KEYFRAME_003) then
    begin
      if SPR_TITLE_ANI = KEYFRAME_001
        then DDSetMFSprFrame(SPR_INTROTEXT_000, -1);
      if SPR_TITLE_ANI = KEYFRAME_002
        then DDSetMFSprFrame(SPR_INTROTEXT_000, -1);
      DDDrawMFSprite(DDBACKS, SPR_INTROTEXT_000, DDBLT_KEYSRC);
    end;

  if RUNSTAGE = RUNSTAGE_MENU then
    begin
      UpdateMainMenu;
      DDDrawMFSprite(DDBACKS, SPR_MENUTEXT_000, DDBLT_KEYSRC);
      DDDrawMFSprite(DDBACKS, SPR_MENUTEXT_001, DDBLT_KEYSRC);
      DDDrawMFSprite(DDBACKS, SPR_MENUTEXT_002, DDBLT_KEYSRC);
    end;

  if RUNSTAGE = RUNSTAGE_INTRO then
    begin
      DDDrawSprite(DDBACKS, SPR_CROP_000, 0);
      DDDrawSprite(DDBACKS, SPR_CROP_001, 0);
    end;

  if RUNSTAGE = RUNSTAGE_OPTION then
    begin
      UpdateOptions;
      RenderOptions;
    end;

  if RUNSTAGE = RUNSTAGE_GAMEINIT then
    begin
      if SPR_TITLE_ANI > -SPR_TITLE_000.SprSrc.Height
        then dec(SPR_TITLE_ANI)
          else DDDrawSprite(DDBACKS, SPR_TEXT1_000, DDBLT_KEYSRC);
      if SPR_ABOUT_ANI < 200 then inc(SPR_ABOUT_ANI);
    end;
end;

// ------------------------------------------------------------------

procedure RenderGameMenu;

begin
  UpdateGameOptions;
  DDSetSpritePos(0, TITLE_POS, SPR_TITLE_000);
  DDDrawSprite(DDBACKS, SPR_GOPTBACK_000, 0);
  DDDrawSprite(DDBACKS, SPR_TITLE_000, DDBLT_KEYSRC);
  RenderOptions;
  DDDrawMFSprite(DDBACKS, SPR_MENUTEXT_002, DDBLT_KEYSRC);
end;

// ------------------------------------------------------------------

procedure IncPlayerScore(_inc: longint);

var _prv : longint;

begin
  _prv := PLAYERSCORE;
  inc(PLAYERSCORE, _inc);
  if ((_prv < 7500)  and (PLAYERSCORE > 7500)) or
     ((_prv < 15000) and (PLAYERSCORE > 15000)) then
    begin
      DMPlay(SND_ITEM_SEG);    
      inc(PLAYERLIFES);
    end;
end;

// ------------------------------------------------------------------

procedure DrawBadGuys;

var _idx : longint;
    _drw : longint;

begin
  _idx := 0;
  _drw := 0;
  if BADGUYS.Count > 0 then
    repeat
      if BADGUYS.BG[_idx].Alive then
        begin
          DDDrawMFSprite(DDBACKS, BADGUYS.BG[_idx].SPR,
                         DDBLT_KEYSRC);
          inc(_drw);
        end;
      inc(_idx);
    until _drw = BADGUYS.Count;
end;

// ------------------------------------------------------------------

procedure DrawNumber(_x, _y, _n: longint; _digits: byte);

var
  _scr : string;
  _cnt : longint;

begin
  if _digits > 0
    then _scr := Format('%.'+IntToStr(_digits)+'d', [_n])
      else _scr := IntToStr(_n);
  for _cnt := 1 to Length(_scr) do
    begin
      DDSetMFSprFrame(SPR_NUMBERS_000, byte(_scr[_cnt])-47);
      DDSetMFSpritePos(_x + ((_cnt-1) * (SPR_NUMBERS_000.SrcWidth + 1)),
                       _y, SPR_NUMBERS_000);
      DDDrawMFSprite(DDBACKS, SPR_NUMBERS_000, DDBLT_KEYSRC);
    end;
end;

// ------------------------------------------------------------------

procedure RenderGame;

var
  _r   : TRect;
  _cnt : longint;

begin
  // Draw Background.
  inc(SPR_BACKGROUND_ANI);
  if (SPR_BACKGROUND_ANI shr 1) >= SPR_BACKGROUND_SRC.Height
    then SPR_BACKGROUND_ANI := 0;
  DDSetSpritePos(0, (SPR_BACKGROUND_ANI shr 1), SPR_BACKGROUND_000);
  DDSetSpritePos(0, (SPR_BACKGROUND_ANI shr 1) -
                     SPR_BACKGROUND_SRC.Height, SPR_BACKGROUND_001);
  DDDrawSprite(DDBACKS, SPR_BACKGROUND_000, 0);
  DDDrawSprite(DDBACKS, SPR_BACKGROUND_001, 0);

  // Draw bad guys
  DrawBadGuys;

  // Draw player shoots
  for _cnt := 0 to NUMPLAYERSHOOTS - 1 do
    if SPR_PLAYSHOT_00X[_cnt].InUse
      then DDDrawMFSprite(DDBACKS, SPR_PLAYSHOT_00X[_cnt].Spr,
                          DDBLT_KEYSRC);

  // Draw player.
  if not PLAYERKILLED then
    begin
      if SPR_PLAYER_ANI < 2
        then inc(SPR_PLAYER_ANI)
          else SPR_PLAYER_ANI := 0;
      if (INVULNERABILITY and 1 = 0)
        then DDDrawMFSprite(DDBACKS, SPR_PLAYER_000, DDBLT_KEYSRC);
    end;

  // Draw player score.
  DDDrawSprite(DDBACKS, SPR_FRAME_000, DDBLT_KEYSRC);
  DrawNumber(54, 187, PLAYERSCORE, 5);
  DrawNumber(213, 187, BLUEITEMCOUNT, 1);
  DrawNumber(240, 187, PLAYERLIFES, 2);
  DrawNumber(276, 187, INVULNERABILITY, 4);

  // Draw stage intro.
  if (SPR_PLANETS_ANI < 333) then
    begin
      inc(SPR_PLANETS_ANI);
      _r := SPR_PLANETS_000.DstRect;
      if SPR_PLANETS_ANI <= 125
        then ScaleRect(SPR_PLANETS_000.DstRect, 125 - SPR_PLANETS_ANI);
      DDDrawSprite(DDBACKS, SPR_STGTITLE_000, DDBLT_KEYSRC);
      DDDrawMFSprite(DDBACKS, SPR_PLANETS_000, 0);
      SPR_PLANETS_000.DstRect := _r;
    end else if (not PLAYINGMUSIC) and (gDSMPMUSIC.MediaControl <> NIL) then
               begin
                 PLAYINGMUSIC := true;
                 UpdateAudioVolume;
                 gDSMPMUSIC.MediaControl.Run;
               end;
end;

// ------------------------------------------------------------------

procedure RenderGameOver;

var _r : TRect;

begin
  inc(SPR_GAMEOVER_ANI);
  _r := SPR_GAMEOVER_000.DstRect;
  if SPR_GAMEOVER_ANI <= 125
    then ScaleRect(SPR_GAMEOVER_000.DstRect, 125 - SPR_GAMEOVER_ANI);
  DDDrawSprite(DDBACKS, SPR_GAMEOVER_000, 0);
  SPR_GAMEOVER_000.DstRect := _r;
  if SPR_GAMEOVER_ANI > 300 then RUNSTAGE := RUNSTAGE_INTROINIT;
end;

// ------------------------------------------------------------------

procedure RenderStageClear;

var _r : TRect;

begin
  inc(SPR_STGCLEAR_ANI);
  _r := SPR_STGCLEAR_000.DstRect;
  if SPR_STGCLEAR_ANI <= 125
    then ScaleRect(SPR_STGCLEAR_000.DstRect, 125 - SPR_STGCLEAR_ANI);
  DDDrawSprite(DDBACKS, SPR_STGCLEAR_000, 0);
  SPR_STGCLEAR_000.DstRect := _r;
end;

// ------------------------------------------------------------------

procedure RenderEndOfGame;

begin
  DDDrawSprite(DDBACKS, SPR_GOPTBACK_000, 0);
  DDDrawSprite(DDBACKS, SPR_TEXT2_000, DDBLT_KEYSRC);
end;

// ------------------------------------------------------------------

function GetBadGuyLife(_clan: longint): longint;

begin
  Result := 0;
  Case _clan of
    0  : Result := 1;  // Tie
    1  : Result := 1;  // Tie ][
    2  : Result := 8;  // Mine
    3  : Result := 25; // Thri-smile
    4  : Result := 10; // TNK-smile
    5  : Result := 2;  // Jump-smile
    6  : Result := 25; // Bi-cannon
    7  : Result := 8;  // Missile-launcher
    8  : Result := 1;  // Missile
    9  : Result := 50; // Piramid
    10 : Result := 3;  // Kamikaze
    11 : Result := 5;  // Container
  end;
end;

// ------------------------------------------------------------------

function GetBadGuyPoints(_clan: longint): longint;

begin
  Result := 0;
  Case _clan of
    0  : Result := 15;  // Tie
    1  : Result := 15;  // Tie ][
    2  : Result := 50;  // Mine
    3  : Result := 250; // Thri-smile
    4  : Result := 75;  // TNK-smile
    5  : Result := 30;  // Jump-smile
    6  : Result := 250; // Bi-cannon
    7  : Result := 75;  // Missile-launcher
    8  : Result := 5;   // Missile
    9  : Result := 500; // Piramid
    10 : Result := 25;  // Kamikaze
    11 : Result := 100; // Container
  end;
end;

// ------------------------------------------------------------------

procedure AddBadGuy(_x, _y, _type, _life, _sbp1: longint);

var _idx : longint;

begin
  _idx := 0;
  if BADGUYS.Count < MAXBADGUYS then
    begin
      // Search free badguy.
      while BADGUYS.BG[_idx].Alive do inc(_idx);
      with  BADGUYS.BG[_idx] do
        begin
          OX := _x;
          OY := _y;
          Clan   := _type;
          Alive  := true;
          Time   := 0;
          Life   := _life;
          SBPar1 := _sbp1;
          Param1 := 0;
          Param2 := 0;
          Case _type of
            0   : Spr := DDGetMFSprite(@SPR_BADGUY00_SRC);
            1   : Spr := DDGetMFSprite(@SPR_BADGUY00_SRC);
            2   : Spr := DDGetMFSprite(@SPR_BADGUY01_SRC);
            3   : Spr := DDGetMFSprite(@SPR_BADGUY02_SRC);
            4   : Spr := DDGetMFSprite(@SPR_BADGUY03_SRC);
            5   : Spr := DDGetMFSprite(@SPR_BADGUY04_SRC);
            6   : Spr := DDGetMFSprite(@SPR_BADGUY05_SRC);
            7   : Spr := DDGetMFSprite(@SPR_BADGUY06_SRC);
            8   : Spr := DDGetMFSprite(@SPR_BGSHOOT4_SRC);
            9   : Spr := DDGetMFSprite(@SPR_BADGUY07_SRC);
            10  : Spr := DDGetMFSprite(@SPR_BADGUY08_SRC);
            11  : Spr := DDGetMFSprite(@SPR_BADGUY09_SRC);
            50  : begin
                    Spr := DDGetMFSprite(@SPR_BADGUY00_SRC);
                    DDSetMFSprFrame(Spr, 3);
                  end;
            51  : Spr := DDGetMFSprite(@SPR_BGSHOOT1_SRC);
            52  : Spr := DDGetMFSprite(@SPR_BGSHOOT2_SRC);
            53  : Spr := DDGetMFSprite(@SPR_BGSHOOT3_SRC);
            54  : Spr := DDGetMFSprite(@SPR_BGSHOOT5_SRC);
            75  : Spr := DDGetMFSprite(@SPR_ITEMS_SRC);
            76  : begin
                    Spr := DDGetMFSprite(@SPR_ITEMS_SRC);
                    DDSetMFSprFrame(Spr, 7);
                  end;
            77  : begin
                    Spr := DDGetMFSprite(@SPR_ITEMS_SRC);
                    DDSetMFSprFrame(Spr, 13);
                  end;
            78  : begin
                    Spr := DDGetMFSprite(@SPR_ITEMS_SRC);
                    DDSetMFSprFrame(Spr, 19);
                  end;
            101 : Spr := DDGetMFSprite(@SPR_EXPLOSION2_SRC);
          end;
          DDSetMFSpritePos(_x, _y, Spr);
          inc(BADGUYS.Count)
        end;
    end;
end;

// ------------------------------------------------------------------

function AddPlayerShoot: boolean;

var _cnt : longint;

begin
  Result := false;
  for _cnt := 0 to NUMPLAYERSHOOTS - 1 do
    if SPR_PLAYSHOT_00X[_cnt].InUse = false then
      begin
        IncPlayerScore(1);
        SPR_PLAYSHOT_00X[_cnt].InUse := true;
        DDSetMFSpritePos(SPR_PLAYER_000.X + 4,
                         SPR_PLAYER_000.Y - (SPR_PLAYER_000.SrcHeight shr 1),
                         SPR_PLAYSHOT_00X[_cnt].Spr);
        Result := true;
        Break;
      end;
end;

// ------------------------------------------------------------------

procedure AnimatePlayerShoots;

var _cnt, _h : longint;

begin
  for _cnt := 0 to NUMPLAYERSHOOTS - 1 do
    with SPR_PLAYSHOT_00X[_cnt] do if InUse then
      begin
        _h := Spr.SrcHeight;
        DDSetMFSpritePos(Spr.X, Spr.Y - (_h shr 1) , Spr);
        InUse := not (Spr.Y < -_h);
      end;
end;

// ------------------------------------------------------------------

procedure InitPlayer;

begin
  NUMPLAYERSHOOTS := DEFPLAYERSHOOTS;
  BLUEITEMCOUNT := 1;
  FASTSHOOT := false;
end;

// ------------------------------------------------------------------

procedure AI00(var _bg: TBADGUY; _ymax: longint);

var _x, _y, _sx : longint;
    _r : single;

begin
  // Position.
  if _bg.Time < _ymax then
    begin
      _r := _bg.Time / 10;
      _bg.Param2 := trunc(cos(_r) * 5);
    end;
  _x := _bg.Spr.X + (_bg.Param2 * _bg.SBPar1);
  _y := _bg.Spr.Y + 1;
  DDSetMFSpritePos(_x, _y, _bg.spr);

  // Shoot.
  _sx := _x + (_bg.Spr.SrcWidth shr 1) - 1;
  if (_bg.Param1 > 0) then inc(_bg.Param1);
  if (_bg.Param1 = TIEFIGHTER_RELOADTIME shr 1) then DDSetMFSprFrame(_bg.Spr, 1);
  if (_bg.Param1 = TIEFIGHTER_RELOADTIME) then _bg.Param1 := 0;
  if (_bg.Time > (TIEFIGHTER_RELOADTIME shr 1)) and
     (not PLAYERKILLED) and (_bg.Param1 = 0) and
     (_y < YLIMIT_FOR_S00TING) and (_sx >= SPR_PLAYER_000.X) and
     (_sx <= SPR_PLAYER_000.X + SPR_PLAYER_000.SrcWidth) then
    begin
      DDSetMFSprFrame(_bg.Spr, 2);
      AddBadGuy(_sx, _y + _bg.Spr.SrcHeight, 50, 0, 0);
      DMPlay(SND_BGSHOOT1_SEG);
      inc(_bg.Param1);
    end;

  // End of life.
  if (_y > 200) or (_x < -50) or (_x > 370) then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI01(var _bg: TBADGUY; _mod: longint);

begin
  if _bg.Time mod _mod = 0 then DDSetMFSprFrame(_bg.Spr, -1);
  if (_bg.Time > _mod) and (_bg.Spr.CellNum = 1)
    then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI02(var _bg: TBADGUY; _mod: longint);

begin
  if (_bg.Time mod _mod = 0) and (_bg.Spr.CellNum < _bg.Spr.SprGrdSrc.Grid.Count)
    then DDSetMFSprFrame(_bg.Spr, -1);
  if (_bg.Time = RESURRECTIONTIME) then
    begin
      dec(PLAYERLIFES);
      if PLAYERLIFES > 0 then
        begin
          InitPlayer;
          PLAYERKILLED := false;
          INVULNERABILITY := INVULNERABILITYTIME;
        end else RUNSTAGE := RUNSTAGE_INITGOVER;
      _bg.Alive := false;
    end;
end;

// ------------------------------------------------------------------

procedure AI03(var _bg: TBADGUY; _inc: byte);

begin
  DDSetMFSpritePos(_bg.Spr.X, _bg.Spr.Y +
                  (_bg.Spr.SrcHeight shr 2) + _inc, _bg.spr);
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI04(var _bg: TBADGUY);

begin
  if _bg.Time mod 2 = 0
    then DDSetMFSpritePos(_bg.Spr.X, _bg.Spr.Y + 1, _bg.spr);
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI05(var _bg: TBADGUY);

var _t : single;
    _x : longint;
    _y : longint;
    _xo : longint;

begin
  _x := _bg.Spr.X;
  _y := _bg.Spr.Y;

  // Y movement.
  Case _bg.Time of
    100 : begin
            DDSetMFSprFrame(_bg.Spr, 2);
            _bg.Param1 := THRI_SMILE_RELOADTIME;
          end;
    400 : DDSetMFSprFrame(_bg.Spr, 4);
  end;
  if (_bg.Time < 100) or (_bg.Time > 400) then
    begin
      if _bg.Time < 400
        then _t := _bg.Time / 5
          else _t := (_bg.Time - 300) / 5;
      _y := _bg.OY + trunc(10 * (_t) - ((_t * _t) / 4));
    end else if not PLAYERKILLED then begin
               // X Movement.
               _xo := _x + (_bg.Spr.SrcWidth shr 1)
                         - (SPR_PLAYER_000.SrcWidth shr 1);
               if (_bg.Time > 115) then
                 begin
                   if (_xo < PLAYER_X) and (_xo < _bg.OX + _bg.SBPar1 + (_bg.Spr.SrcWidth shr 1)) then inc(_x) else
                     if (_xo > PLAYER_X) and (_xo > _bg.OX - _bg.SBPar1) then dec(_x);
                 end;

               // Shooting
               if _bg.Param1 = 5 then DDSetMFSprFrame(_bg.Spr, 2);
               if _bg.Param1 = THRI_SMILE_RELOADTIME then
                 begin
                   DDSetMFSprFrame(_bg.Spr, 3);
                   AddBadGuy(_x,  _y + _bg.Spr.SrcHeight - 2, 51, 0, -1);
                   AddBadGuy(_xo + 6, _y + _bg.Spr.SrcHeight - 2, 51, 0,  0);
                   AddBadGuy(_x + _bg.Spr.SrcWidth - 6,
                             _y + _bg.Spr.SrcHeight - 2, 51, 0,  1);
                   DMPlay(SND_BGSHOOT2_SEG);
                   _bg.Param1 := 0;
                 end else inc(_bg.Param1);
             end else if _bg.Spr.CellNum = 3 then DDSetMFSprFrame(_bg.Spr, 2);
  DDSetMFSpritePos(_x, _y, _bg.spr);

  // End of life.
  if _bg.Spr.Y < -100 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI06(var _bg: TBADGUY);

begin
  if _bg.Time mod 2 = 0
    then DDSetMFSprFrame(_bg.Spr, -1);
  DDSetMFSpritePos(_bg.Spr.X + _bg.SBPar1, _bg.Spr.Y + 1, _bg.spr);
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI07(var _bg: TBADGUY);

var _s : extended;
    _y : longint;

begin
  // Get angle.
  _y := PLAYER_Y - _bg.Spr.Y;
  if (_y <> 0) and (not PLAYERKILLED)
   and (_bg.Spr.Y < YLIMIT_FOR_S00TING) then
    begin
      _s := Hypot(PLAYER_X - _bg.Spr.X, PLAYER_Y - _bg.Spr.Y) / (_y);
      if (_s <> 0) then
        begin
          _s := RadToDeg(ArcCos(1/_s));
          if _bg.Spr.X > PLAYER_X then _s := _s * -1;
          if (_s > - 20) and (_s < 20) then DDSetMFSprFrame(_bg.Spr, 1);
          if (_s <= - 20) then DDSetMFSprFrame(_bg.Spr, 3);
          if (_s >= 20) then DDSetMFSprFrame(_bg.Spr, 2);
        end;

      // Shooting
      if (_bg.Param1 = TNK_SMILE_RELOADTIME) then
        begin
          _y := _bg.Spr.y + _bg.Spr.SrcHeight - 2;
          Case _bg.Spr.CellNum of
            2 : AddBadGuy(_bg.Spr.X + _bg.Spr.SrcWidth, _y, 51, 0, 1);
            3 : AddBadGuy(_bg.Spr.X, _y, 51, 0, -1);
            else AddBadGuy(_bg.Spr.X + 6, _y, 51, 0, 0);
          end;
          DMPlay(SND_BGSHOOT2_SEG);
          _bg.Param1 := 0;
        end else inc(_bg.Param1);
    end;

  // Move
  if _bg.Time mod 2 = 0
    then DDSetMFSpritePos(_bg.Spr.X, _bg.Spr.Y + 1, _bg.spr);

  // End of life
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI08(var _bg: TBADGUY);

var _t : single;
    _y : longint;

begin
  // Movement
  _t := _bg.Time / 5;
  _y := _bg.OY + trunc(-9 * (_t) + ((_t * _t) / 4));
  if (_bg.Time = 40) or (_bg.Time = 60)
    then DDSetMFSprFrame(_bg.Spr, -1);

  // Shooting
  if (not PLAYERKILLED) and (_bg.Spr.Y < YLIMIT_FOR_S00TING) then
    if (_bg.Param1 = JUMP_SMILE_RELOADTIME) then
      begin
        DMPlay(SND_BGSHOOT1_SEG);
        AddBadGuy(_bg.Spr.X + 7, _bg.Spr.Y + _bg.Spr.SrcHeight - 4,
                  52, 0, 0);
        _bg.Param1 := 0;
      end else inc(_bg.Param1);

  // Move
  DDSetMFSpritePos(_bg.Spr.X + _bg.SBPar1, _y, _bg.spr);

  // End of life
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI09(var _bg: TBADGUY);

const IA09_EOA = 380;

var _t : single;
    _x : longint;
    _y : longint;
    _xo : longint;
    _yo : longint;

begin
  // Movement
  _x := _bg.Spr.X;
  _y := _bg.Spr.Y;
  _yo := _bg.Spr.Y + _bg.Spr.SrcHeight - 10;
  case _bg.Time of
     60 : DDSetMFSprFrame(_bg.Spr, 6);
    100 : _bg.OY := _y;
    IA09_EOA : DDSetMFSprFrame(_bg.Spr, 1);
    else begin
           if (_bg.Time < 100) then
             begin
               _t := _bg.Time / 5;
               _y := _bg.OY + trunc(10 * (_t) - ((_t * _t) / 4));
             end;
           if (_bg.Time > IA09_EOA) then
             begin
               _t := (_bg.Time - IA09_EOA) / 5;
               _y := _bg.OY + trunc((_t * _t) / 4);
             end;
           if (_bg.Time > 100) and (_bg.Time < IA09_EOA) then
             begin
               _xo := _x + (_bg.Spr.SrcWidth shr 1)
                         - (SPR_PLAYER_000.SrcWidth shr 1);
               if (_bg.Time > 115) then
                 begin
                   if (_xo < PLAYER_X) and (_xo < _bg.OX + _bg.SBPar1 + (_bg.Spr.SrcWidth shr 1)) then inc(_x) else
                     if (_xo > PLAYER_X) and (_xo > _bg.OX - _bg.SBPar1) then dec(_x);

                   // Cannon animation
                   if (_bg.Spr.CellNum <> 6) and (_bg.Time mod 4 = 0) then
                     if (_bg.Spr.CellNum = 5) or (_bg.Spr.CellNum = 10)
                       then DDSetMFSprFrame(_bg.Spr, 6)
                         else DDSetMFSprFrame(_bg.Spr, -1);

                   // Shooting
                   _xo := _x + _bg.Spr.SrcWidth - 6;
                   if (not PLAYERKILLED) then
                     if _bg.Param1 = BI_CANNON_RELOADTIME then
                       begin
                         DMPlay(SND_BGSHOOT1_SEG);
                         if _bg.Param2 = 0 then
                           begin
                             AddBadGuy(_x + 2, _yo, 53, 0, 0);
                             DDSetMFSprFrame(_bg.Spr, 2);
                             _bg.Param2 := 1;
                           end else begin
                                      AddBadGuy(_xo, _yo, 53, 0, 0);
                                      DDSetMFSprFrame(_bg.Spr, 7);
                                      _bg.Param2 := 0;
                                    end;
                         _bg.Param1 := 0;
                       end else inc(_bg.Param1);
                 end;
             end;
         end;
  end;

  // Move
  DDSetMFSpritePos(_x, _y, _bg.spr);

  // End of life
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI10(var _bg: TBADGUY; _mod, _dec: longint);

begin
  if (_bg.Time mod _mod = 0) then DDSetMFSprFrame(_bg.Spr, -1);
  DDSetMFSpritePos(_bg.Spr.X, _bg.Spr.Y +
                  (_bg.Spr.SrcHeight shr 2) - _dec, _bg.spr);

  // End of life
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI11(var _bg: TBADGUY);

var _yo : longint;

begin
  // Movement
  if _bg.Time mod 2 = 0
    then DDSetMFSpritePos(_bg.Spr.X, _bg.Spr.Y + 1, _bg.spr);

  // Shooting
  if (_bg.Param1 = 5) then DDSetMFSprFrame(_bg.Spr, 1);
  if (_bg.Param1 = MLAUNCHER_RELOADTIME) then
    begin
      if (not PLAYERKILLED) and (_bg.Spr.Y < YLIMIT_FOR_S00TING) then
        begin
          _yo := _bg.Spr.Y + _bg.Spr.SrcHeight;
          DMPlay(SND_MLAUNCH_SEG);
          if _bg.SBPar1 = 0 then
            begin
              AddBadGuy(_bg.Spr.X, _yo, 8, 0, 0);
              DDSetMFSprFrame(_bg.Spr, 2);
              _bg.SBPar1 := 1;
            end else begin
                       AddBadGuy(_bg.Spr.X + _bg.Spr.SrcWidth - 7, _yo, 8, 0, 0);
                       DDSetMFSprFrame(_bg.Spr, 3);
                       _bg.SBPar1 := 0;
                     end;
        end;
      _bg.Param1 := 0;
    end else inc(_bg.Param1);

  // End of life
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI12(var _bg: TBADGUY);

var _xo : longint;
    _xm : longint;
    _x : longint;

begin
  // Movement
  _x := _bg.Spr.X;
  _xo := PLAYER_X + (SPR_PLAYER_000.SrcWidth shr 1);
  _xm := _x + (_bg.Spr.SrcWidth shr 1);
  if _xm > _xo then dec(_bg.Param1) else
    if _xm < _xo then inc(_bg.Param1);
  if _bg.Time mod 3 = 0 then _x := _bg.Spr.X + _bg.Param1 div 25;
  DDSetMFSpritePos(_x, _bg.Spr.Y + 2, _bg.spr);

  // Animation
  if _bg.Time mod 2 = 0 then DDSetMFSprFrame(_bg.Spr, -1);

  // End of life
  if _bg.Spr.Y > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI13(var _bg: TBADGUY);

const IA13_BOA = 100;
      IA13_EOA = 550;

var _t : single;
    _y : longint;
    _x : longint;
    _xo : longint;
    _xf : longint;
    _yo : longint;
    _fire : longint;

begin
  _fire := PIRAMID_RELOADTIME + PIRAMID_FIRETIME;

  Case _bg.Time of
    IA13_BOA - 30 : DDSetMFSprFrame(_bg.Spr, 2);
    IA13_EOA + 1 : DDSetMFSprFrame(_bg.Spr, 3);
  end;

  _x := _bg.Spr.X;
  _y := _bg.Spr.Y;
  _yo := _y + _bg.Spr.SrcHeight;
  if (_bg.Time < IA13_BOA) or (_bg.Time > IA13_EOA) then
    begin
      // Movement
      if _bg.Time < IA13_EOA
        then _t := _bg.Time / 5
          else _t := (_bg.Time - (IA13_EOA - IA13_BOA)) / 5;
      _y := _bg.OY + trunc(8 * (_t) - ((_t * _t) / 5));
    end else begin
               // Aim
               _xf := _x + (_bg.Spr.SrcWidth shr 1);
               if not PLAYERKILLED then
                 begin
                   _xo := PLAYER_X + (SPR_PLAYER_000.SrcWidth shr 1);
                   if _xo < _xf then dec(_x) else
                     if _xo > _xf then inc(_x);
                 end;

               inc(_bg.Param1);

               // Second. weapon
               if (_bg.Param1 mod PIRAMID_RELOADTIME2 = 0) then
                 begin
                   AddBadGuy(_x + 16, _yo-6, 51, 0,  0);
                   AddBadGuy(_x + _bg.Spr.SrcWidth - 22,
                             _yo-6, 51, 0,  0);
                   DMPlay(SND_BGSHOOT2_SEG);                             
                 end;

               // Primary weapon (big laser)
               if (_bg.Param1 < PIRAMID_RELOADTIME) then
                 begin
                   // Animation
                   if _bg.Time mod 2 = 0 then
                     if _bg.Spr.CellNum > 3 then
                       begin
                         if _bg.Spr.CellNum < 6
                           then DDSetMFSprFrame(_bg.Spr, -1)
                             else DDSetMFSprFrame(_bg.Spr, 4);
                       end else DDSetMFSprFrame(_bg.Spr, 4);
                 end else begin
                            _xf := _x + (_bg.Spr.SrcWidth shr 1) - 2;
                            DMPlay(SND_BIGLASER_SEG);
                            AddBadGuy(_xf, _yo, 54, 0, 0);
                            AddBadGuy(_xf, _yo + 29, 54, 0, 0);
                            AddBadGuy(_xf, _yo + 58, 54, 0, 0);
                            AddBadGuy(_xf, _yo + 87, 54, 0, 0);
                          end;
               if (_bg.Param1 = _fire) then
                 begin
                   _bg.Param1 := 0;
                 end;
             end;

  // Move
  DDSetMFSpritePos(_x, _y, _bg.spr);

  // End of life
  if _bg.Spr.Y < -100 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI14(var _bg: TBADGUY);

begin
  // End of life
  if _bg.Time > 0 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI15(var _bg: TBADGUY; _mod, _speed: longint);

begin
  if _bg.Time mod _mod = 0 then DDSetMFSprFrame(_bg.Spr, -1);
  DDSetMFSpritePos(_bg.Spr.X, _bg.Spr.Y + _speed, _bg.spr);

  // End of life
  if _bg.Time > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AI16(var _bg: TBADGUY; _from, _to, _mod: longint);

begin
  if _bg.Time mod _mod = 0 then
    if _bg.Spr.CellNum = _to
      then DDSetMFSprFrame(_bg.Spr, _from)
        else DDSetMFSprFrame(_bg.Spr, -1);
  DDSetMFSpritePos(_bg.Spr.X, _bg.Spr.Y + 1, _bg.spr);

  // End of life
  if _bg.Time > 200 then _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure AnimateBadGuys;

var _idx : longint;
    _ani : longint;
    _out : longint;

begin
  _idx := 0;
  _ani := 0;
  _out := 0;
  if BADGUYS.Count > 0 then
    repeat
      if (_idx > MAXBADGUYS-1) then _idx := 0;
      if BADGUYS.BG[_idx].Alive then
        begin
          Case BADGUYS.BG[_idx].Clan of
            0   : AI00(BADGUYS.BG[_idx], 70);   // Tie-smile I
            1   : AI00(BADGUYS.BG[_idx], 1000); // Tie-smile II
            2   : AI04(BADGUYS.BG[_idx]);       // Mine
            3   : AI05(BADGUYS.BG[_idx]);       // Thri-smile
            4   : AI07(BADGUYS.BG[_idx]);       // TNK-smile
            5   : AI08(BADGUYS.BG[_idx]);       // Jump-smile
            6   : AI09(BADGUYS.BG[_idx]);       // Bi-cannon
            7   : AI11(BADGUYS.BG[_idx]);       // Missile launcher
            8   : AI12(BADGUYS.BG[_idx]);       // Missile
            9   : AI13(BADGUYS.BG[_idx]);       // Piramid
            10  : AI15(BADGUYS.BG[_idx], 2, 2); // Kamikaze
            11  : AI04(BADGUYS.BG[_idx]);       // Container
            50  : AI03(BADGUYS.BG[_idx], 0);    // Tie-smile shoot
            51  : AI06(BADGUYS.BG[_idx]);       // Thri/TNK-smile shoot
            52  : AI03(BADGUYS.BG[_idx], 1);    // Jump-smile shoot
            53  : AI10(BADGUYS.BG[_idx], 2, 1); // Bi-cannon shoot
            54  : AI14(BADGUYS.BG[_idx]);       // Piramid shoot
            75  : AI16(BADGUYS.BG[_idx],  1,  6, 2); // Life
            76  : AI16(BADGUYS.BG[_idx],  7, 12, 2); // Invulnerability
            77  : AI16(BADGUYS.BG[_idx], 13, 18, 2); // Inc # shoots
            78  : AI16(BADGUYS.BG[_idx], 19, 24, 2); // Points
            100 : AI01(BADGUYS.BG[_idx], 3);    // Explosion 1
            101 : AI02(BADGUYS.BG[_idx], 4);    // Explosion 2 (player)
            102 : AI01(BADGUYS.BG[_idx], 3);    // Explosion 3
            103 : AI01(BADGUYS.BG[_idx], 3);    // Explosion 4
            104 : AI01(BADGUYS.BG[_idx], 4);    // Container Expl.
          end;
          inc(BADGUYS.BG[_idx].Time); // Increase time of life.
          if not BADGUYS.BG[_idx].Alive
            then inc(_out);
          inc(_ani);
        end;
      inc(_idx);
    until _ani = BADGUYS.Count;
  dec(BADGUYS.Count, _out);
end;

// ------------------------------------------------------------------

function DestroyBadGuy(var _bg: TBADGUY): boolean;

var _x, _y: longint;

begin
  Result := true;
  IncPlayerScore(GetBadGuyPoints(_bg.Clan));
  if (_bg.Clan < 50) then
    begin
      Result := false;
      _x := _bg.Spr.X;
      _y := _bg.Spr.Y;
      Case _bg.Clan of
        2: begin
             DMPlay(SND_EXPLOSION3_SEG);
             _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION3_SRC);
             dec(_x, 3); dec(_y, 3);
             _bg.Clan := 102;
           end;
        3: begin
             DMPlay(SND_EXPLOSION3_SEG);
             _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION4_SRC);
             dec(_x, 6); dec(_y, 5);
             _bg.Clan := 103;
           end;
        4: begin
             DMPlay(SND_EXPLOSION3_SEG);
             _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION3_SRC);
             _bg.Clan := 102;
           end;
        5: begin
             DMPlay(SND_EXPLOSION3_SEG);
             _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION1_SRC);
             inc(_x); inc(_y, 6);
             _bg.Clan := 100;
           end;
        6: begin
             DMPlay(SND_EXPLOSION3_SEG);
             _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION4_SRC);
             dec(_x, 12); dec(_y, 3);
             _bg.Clan := 103;
           end;
        7: begin
             DMPlay(SND_EXPLOSION3_SEG);
             _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION3_SRC);
             inc(_x, 4); dec(_y, 2);
             _bg.Clan := 102;
           end;
        8: begin
             DMPlay(SND_EXPLOSION3_SEG);
             _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION1_SRC);
             dec(_x, 6); inc(_y, 5);
             _bg.Clan := 100;
           end;
        9: begin
             DMPlay(SND_EXPLOSION3_SEG);
             _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION4_SRC);
             inc(_x, 28); inc(_y, 18);
             _bg.Clan := 103;
           end;
        10: begin
              DMPlay(SND_EXPLOSION3_SEG);
              _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION1_SRC);
              dec(_x, 4); inc(_y, 7);
              _bg.Clan := 100;
            end;
        11: begin
              DMPlay(SND_EXPLOSION3_SEG);
              DDSetMFSprFrame(_bg.Spr, 2);
              _bg.Clan := 104;
              case Random(4) of
                0 : AddBadGuy(_bg.Spr.X, _bg.Spr.Y, 75, 0, 0);
                1 : AddBadGuy(_bg.Spr.X, _bg.Spr.Y, 76, 0, 0);
                2 : AddBadGuy(_bg.Spr.X, _bg.Spr.Y, 77, 0, 0);
                3 : AddBadGuy(_bg.Spr.X, _bg.Spr.Y, 78, 0, 0);
              end;
            end;
        else begin
               DMPlay(SND_EXPLOSION3_SEG);
               _bg.Spr := DDGetMFSprite(@SPR_EXPLOSION1_SRC);
               _bg.Clan := 100;
             end;
      end;
      DDSetMFSpritePos(_x, _y, _bg.Spr);
      _bg.Time := 0;
    end else _bg.Alive := false;
end;

// ------------------------------------------------------------------

procedure CheckBadGuysCollisions;

var _idx : longint;
    _chk : longint;
    _cnt : longint;
    _out : longint;
    _crash : boolean;

begin
  _out := 0;
  _idx := 0;
  _chk := 0;
  if BADGUYS.Count > 0 then
    repeat
      if (BADGUYS.BG[_idx].Alive) and (BADGUYS.BG[_idx].Clan < 100) then
        begin
          _crash := false;

          // Check collisions with user lasers
          if (BADGUYS.BG[_idx].Clan < 50) then
            for _cnt := 0 to NUMPLAYERSHOOTS - 1 do
              if (SPR_PLAYSHOT_00X[_cnt].InUse) and
                 (RectCollision(BADGUYS.BG[_idx].Spr.DstRect,
                                SPR_PLAYSHOT_00X[_cnt].Spr.DstRect)) then
                begin
                  DMPlay(SND_EXPLOSION1_SEG);
                  SPR_PLAYSHOT_00X[_cnt].InUse := false;
                  _crash := true;
                  break;
                end;

          // Check collisions with user
          if (not _crash) and (not PLAYERKILLED) and
             (RectCollision(BADGUYS.BG[_idx].Spr.DstRect,
                            SPR_PLAYER_000.DstRect)) then
            if (BADGUYS.BG[_idx].Clan < 75) then
              begin
                if (INVULNERABILITY = 0) then
                  begin
                    PLAYERKILLED := true;
                    DMPlay(SND_EXPLOSION2_SEG);
                    AddBadGuy(SPR_PLAYER_000.X - 16,
                              SPR_PLAYER_000.Y, 101, 0, 0);
                    BADGUYS.BG[_idx].Life := 0;
                    _crash := true;
                  end;
              end else begin
                         // Got Item
                         DMPlay(SND_ITEM_SEG);
                         Case BADGUYS.BG[_idx].Clan of
                           75 : if PLAYERLIFES < 99
                                  then inc(PLAYERLIFES); // Life
                           76 : if INVULNERABILITY < 9000
                                  then inc(INVULNERABILITY, 500); // Inv
                           77 : if NUMPLAYERSHOOTS < MAXPLAYERSHOOTS then
                                  begin
                                    inc(NUMPLAYERSHOOTS);
                                    inc(BLUEITEMCOUNT);
                                  end else
                                  if not FASTSHOOT then
                                    begin
                                      FASTSHOOT := true;
                                      inc(BLUEITEMCOUNT);
                                    end else IncPlayerScore(1250);
                           78 : IncPlayerScore(2500);
                         end;
                         BADGUYS.BG[_idx].Alive := false;
                         inc(_out);
                       end;

          // Bad guy destroyed.
          if (_crash) then with BADGUYS.BG[_idx] do
            begin
              dec(Life);
              if Life <= 0 then
                if DestroyBadGuy(BADGUYS.BG[_idx])
                  then inc(_out);
            end;
          inc(_chk);
        end;
      inc(_idx);
    until _chk = BADGUYS.Count;
  dec(BADGUYS.Count, _out);
end;

// ------------------------------------------------------------------

procedure ProcessStoryBoard;

var _size  : longint;
    _event : array[0..4] of longint;

begin
  if STORYBOARDCUR <> NIL then
    begin
      _size := SizeOf(_event);
      Move(STORYBOARDCUR^, _event, _size);
      while (_event[0] < STAGETIME) and (STORYBOARDCUR <> NIL) do
        begin
          if _event[3] <> $FFFFFF then
            begin
              AddBadGuy(_event[1], _event[2], _event[3],
                        GetBadGuyLife(_event[3]), _event[4]);
              inc(longint(STORYBOARDCUR), _size);
              Move(STORYBOARDCUR^, _event, _size);
            end else begin
                       STORYBOARDCUR := NIL;
                       RUNSTAGE := RUNSTAGE_STGCINIT;
                     end;
        end;
    end;
end;

// ------------------------------------------------------------------

procedure RefreshInputs(var _up, _down, _left, _right,
                            _shot, _esc, _sF11, _sF12 : longint;
                        var _fleft, _fright : boolean);

var _time : dWord;
    _joy  : TDIJoyState2;
    _fst  : longint;

begin
  _up    := 0;
  _down  := 0;
  _left  := 0;
  _right := 0;
  _shot  := 0;
  _esc   := 0;

  // keyboard.
  _time := GetTickCount;
  if DIGetKeyboardState then
    begin
      // Check for fast shot.
      if (RUNSTAGE = RUNSTAGE_GAME) and (FASTSHOOT)
        then _fst := KEYTIMEOUT shr 1 else _fst := KEYTIMEOUT;
      // Read keys.
      _up     := TimedInputRefresh(KTI_UP,    _time, KEYTIMEOUT, DIKeyDown(DIK_UP));
      _down   := TimedInputRefresh(KTI_DOWN,  _time, KEYTIMEOUT, DIKeyDown(DIK_DOWN));
      _left   := TimedInputRefresh(KTI_LEFT,  _time, KEYTIMEOUT, DIKeyDown(DIK_LEFT));
      _right  := TimedInputRefresh(KTI_RIGHT, _time, KEYTIMEOUT, DIKeyDown(DIK_RIGHT));
      _shot   := TimedInputRefresh(KTI_SHOT,  _time, _fst, (DIKeyDown(DIK_SPACE)) or (DIKeyDown(DIK_RETURN)));
      _esc    := TimedInputRefresh(KTI_ESC,   _time, KEYTIMEOUT, DIKeyDown(DIK_ESCAPE));
      _sF11   := TimedInputRefresh(KTI_SF11,  _time, KEYTIMEOUT, (DIKeyDown(DIK_LSHIFT)) and (DIKeyDown(DIK_F11)));
      _sF12   := TimedInputRefresh(KTI_SF12,  _time, KEYTIMEOUT, (DIKeyDown(DIK_LSHIFT)) and (DIKeyDown(DIK_F12)));
      _fleft  := DIKeyDown(DIK_LEFT);
      _fright := DIKeyDown(DIK_RIGHT);
    end;

  // joystick, gamepad,...
  if (JOYSTICKAVAILABLE) and (DIGetJoystickState(_joy)) then
    begin
      // Check for fast shot.
      if (RUNSTAGE = RUNSTAGE_GAME) and (FASTSHOOT)
        then _fst := JOYTIMEOUT shr 1 else _fst := JOYTIMEOUT;
      // Read input.
      if _up    = 0 then _up    := TimedInputRefresh(JTI_UP,
                                     _time, JOYTIMEOUT, (_joy.lY < - 20000));
      if _down  = 0 then _down  := TimedInputRefresh(JTI_DOWN,
                                     _time, JOYTIMEOUT, (_joy.lY > 20000));
      if _left  = 0 then _left  := TimedInputRefresh(JTI_LEFT,
                                     _time, JOYTIMEOUT, (_joy.lX < - 20000));
      if _right = 0 then _right := TimedInputRefresh(JTI_RIGHT,
                                     _time, JOYTIMEOUT, (_joy.lX > 20000));
      if _shot  = 0 then _shot  := TimedInputRefresh(JTI_SHOT,
                                     _time, _fst, (_joy.rgbButtons[0] > 0));
      if _esc   = 0 then _esc   := TimedInputRefresh(JTI_ESC,
                                     _time, JOYTIMEOUT, (_joy.rgbButtons[1] > 0));
      if (not _fleft)  then _fleft  := (_joy.lX < - 20000);
      if (not _fright) then _fright := (_joy.lX > 20000);
    end;
end;

// ------------------------------------------------------------------

procedure RenderFrame(_hWnd: hWnd);

var _up, _down, _left, _right, _shot, _esc, _sF11, _sF12 : longint;
    _fleft, _fright : boolean;

begin
  if DDBACKS.IsLost <> 0 then
    begin
      DD7.RestoreAllSurfaces;
      // Reacquire input devices.
      if INIT_FULLSCREEN then DIMouseControl(true);
      DIKeyboardControl(true);
      if JOYSTICKAVAILABLE then DIJoystickControl(true);
    end;

  // FPS control.
  while GetTickCount - FRAMECONTROL < FRAME_TIME do sleep(0);
  FRAMECONTROL := GetTickCount;

  // Muscic loop test.
  inc(MUSICLOOPTESTCNT);
  if MUSICLOOPTESTCNT > MUSICLOOPTESTIN then
    begin
      MUSICLOOPTESTCNT := 0;
      MusicLoopTest;
    end;

  // Run Stages
  Case RUNSTAGE of
    RUNSTAGE_INTROINIT : begin
                           InitializeIntro;
                           RUNSTAGE := RUNSTAGE_INTRO;
                         end;
    RUNSTAGE_INTRO : begin
                       RenderIntro;
                       RefreshInputs(_up, _down, _left, _right, _shot,
                                     _esc, _sF11, _sF12, _fleft, _fright);
                       if (_esc > 0) then
                         begin
                           RUNSTAGE := RUNSTAGE_MENUINIT;
                           ResetTimedInputs;
                         end;
                     end;
    RUNSTAGE_MENUINIT : begin
                          InitializeMainMenu;
                          RUNSTAGE := RUNSTAGE_MENU;
                        end;
    RUNSTAGE_MENU : begin
                      RenderIntro;
                      RefreshInputs(_up, _down, _left, _right, _shot,
                                     _esc, _sF11, _sF12, _fleft, _fright);
                      if (_up > 0)
                        then MinMax(MENUSELECTION, -1, 0, 2, true);
                      if (_down > 0)
                        then MinMax(MENUSELECTION,  1, 0, 2, true);

                      if (_shot > 0) then
                        begin
                          DMPlay(SND_MENUSEL_SEG);
                          case MENUSELECTION of
                            0 : begin
                                  STAGENO := 1;
                                  PLAYERSCORE := 0;
                                  PLAYERLIFES := 3;
                                  InitPlayer;
                                  RUNSTAGE := RUNSTAGE_GAMEINIT;
                                end;
                            1 : RUNSTAGE := RUNSTAGE_OPTIONINIT;
                            2 : Halt;
                          end;
                        end;
                    end;
    RUNSTAGE_OPTIONINIT : begin
                            InitializeMainMenu;
                            RUNSTAGE := RUNSTAGE_OPTION;
                          end;
    RUNSTAGE_OPTION : begin
                        RenderIntro;
                        RefreshInputs(_up, _down, _left, _right, _shot,
                                     _esc, _sF11, _sF12, _fleft, _fright);

                        if (_right > 0) then
                          begin
                            if OPTIONSELECTION = 0
                              then MinMax(MUSICVOLUME,  1, 0, MAXVOLUME, true);
                            if OPTIONSELECTION = 1
                              then MinMax(SOUNDVOLUME,  1, 0, MAXVOLUME, true);
                            UpdateAudioVolume;
                          end;

                        if (_left > 0) then
                          begin
                            if OPTIONSELECTION = 0
                              then MinMax(MUSICVOLUME, -1, 0, MAXVOLUME, true);
                            if OPTIONSELECTION = 1
                              then MinMax(SOUNDVOLUME, -1, 0, MAXVOLUME, true);
                            UpdateAudioVolume;
                          end;

                        if (_up > 0)
                          then MinMax(OPTIONSELECTION, -1, 0, 2, true);

                        if (_down > 0)
                          then MinMax(OPTIONSELECTION,  1, 0, 2, true);

                        if (_shot > 0) then
                          begin
                            DMPlay(SND_MENUSEL_SEG);
                            if OPTIONSELECTION = 2
                              then RUNSTAGE := RUNSTAGE_MENUINIT;
                          end;
                      end;
    RUNSTAGE_GAMEINIT : begin
                          RenderIntro;
                          RefreshInputs(_up, _down, _left, _right, _shot,
                                        _esc, _sF11, _sF12, _fleft, _fright);
                          if (_shot > 0) then RUNSTAGE := RUNSTAGE_STAGEBEGIN;
                        end;
    RUNSTAGE_STAGEBEGIN : begin
                            gDSMPMUSIC.MediaControl.Stop;
                            InitializeStage;
                            STAGETIME := 0;
                            PLAYINGMUSIC := false;
                            PLAYERKILLED := false;
                            INVULNERABILITY := 0;
                            case STAGENO of
                              1: STORYBOARDCUR := @STAGE_1;
                              2: STORYBOARDCUR := @STAGE_2;
                              3: STORYBOARDCUR := @STAGE_3;
                              4: STORYBOARDCUR := @STAGE_4;
                              5: begin
                                   gDSMPMUSIC.MediaControl.Stop;
                                   dsmpLoadFile(gDSMPMUSIC, GAMEMUSICPATH + MUSIC_END);
                                   gDSMPMUSIC.MediaControl.Run;
                                   RUNSTAGE := RUNSTAGE_EOGAME;
                                   exit;
                                 end;
                            end;
                            FillChar(BADGUYS, SizeOf(BADGUYS), 0);
                            DMPlay(SND_STAGEINI_SEG);
                            RUNSTAGE := RUNSTAGE_GAME;
                          end;
    RUNSTAGE_EOGAME : begin
                        RenderEndOfGame;
                        RefreshInputs(_up, _down, _left, _right, _shot,
                                      _esc, _sF11, _sF12, _fleft, _fright);
                        if (_shot > 0) then
                          begin
                            gDSMPMUSIC.MediaControl.Stop;
                            RUNSTAGE := RUNSTAGE_INTROINIT;
                          end;
                      end;
    RUNSTAGE_GAME : begin
                      RenderGame;
                      inc(STAGETIME);
                      if INVULNERABILITY > 0 then dec(INVULNERABILITY);
                      AnimateBadGuys;
                      AnimatePlayerShoots;
                      CheckBadGuysCollisions;
                      ProcessStoryBoard;

                      // Input
                      RefreshInputs(_up, _down, _left, _right, _shot,
                                    _esc, _sF11, _sF12, _fleft, _fright);

                      // Player input ----------
                      if (not PLAYERKILLED) then
                        begin
                          DDSetMFSprFrame(SPR_PLAYER_000,
                                          1 + SPR_PLAYER_ANI);

                          if (_shot > 0) then
                            if AddPlayerShoot
                              then DMPlay(SND_PLAYSHOT_SEG);

                          if (_fleft) and (PLAYER_X > 3) then
                            begin
                              dec(PLAYER_X, 2);
                              DDSetMFSprFrame(SPR_PLAYER_000,
                                              7 + SPR_PLAYER_ANI);
                            end;

                          if (_fright) and (PLAYER_X < 317 - SPR_PLAYER_000.SrcWidth) then
                            begin
                              inc(PLAYER_X, 2);
                              DDSetMFSprFrame(SPR_PLAYER_000,
                                              4 + SPR_PLAYER_ANI);
                            end;

                          DDSetMFSpritePos(PLAYER_X, PLAYER_Y, SPR_PLAYER_000);
                        end;
                      // Player input ----------

                      // Special input ---------
                      if (_esc > 0) then
                        begin
                          DMPlay(SND_MENUSEL_SEG);
                          GAMEOPTSELECTION := 0;
                          RUNSTAGE := RUNSTAGE_GAMEMENU;
                        end;
                      if (_sF11 > 0) and (INVULNERABILITY < 9000)
                        then inc(INVULNERABILITY, 1000);
                      if (_sF12 > 0) and (PLAYERLIFES < 99)
                        then inc(PLAYERLIFES);
                      // Special input ---------
                    end;
    RUNSTAGE_GAMEMENU : begin
                          RefreshInputs(_up, _down, _left, _right, _shot,
                                        _esc, _sF11, _sF12, _fleft, _fright);

                          if (_up > 0)
                            then MinMax(GAMEOPTSELECTION, -1, 0, 3, true);

                          if (_down > 0)
                            then MinMax(GAMEOPTSELECTION,  1, 0, 3, true);

                          if (_right > 0) then
                            begin
                              if GAMEOPTSELECTION = 0
                                then MinMax(MUSICVOLUME,  1, 0, MAXVOLUME, true);
                              if GAMEOPTSELECTION = 1
                                then MinMax(SOUNDVOLUME,  1, 0, MAXVOLUME, true);
                              UpdateAudioVolume;
                            end;

                          if (_left > 0) then
                            begin
                              if GAMEOPTSELECTION = 0
                                then MinMax(MUSICVOLUME, -1, 0, MAXVOLUME, true);
                              if GAMEOPTSELECTION = 1
                                then MinMax(SOUNDVOLUME, -1, 0, MAXVOLUME, true);
                              UpdateAudioVolume;
                            end;

                          if (_shot > 0) then
                            begin
                              DMPlay(SND_MENUSEL_SEG);
                              case GAMEOPTSELECTION of
                                2 : RUNSTAGE := RUNSTAGE_GAME;
                                3 : RUNSTAGE := RUNSTAGE_INITGOVER;
                              end;
                            end;

                          RenderGameMenu;
                        end;
    RUNSTAGE_INITGOVER : begin
                           SPR_GAMEOVER_000 := DDGetSprite(@SPR_GAMEOVER_SRC);
                           DDSetSpritePos(1, 100 -
                                         (SPR_GAMEOVER_000.SprSrc.Height shr 1),
                                          SPR_GAMEOVER_000);
                           gDSMPMUSIC.MediaControl.Stop;
                           PLAYINGMUSIC := false;
                           SPR_GAMEOVER_ANI := 0;
                           DMPlay(SND_STAGEINI_SEG);
                           RUNSTAGE := RUNSTAGE_GAMEOVER;
                         end;
    RUNSTAGE_GAMEOVER  : RenderGameOver;
    RUNSTAGE_STGCINIT  : begin
                           gDSMPMUSIC.MediaControl.Stop;
                           PLAYINGMUSIC := false;

                           SPR_STGCLEAR_000 := DDGetSprite(@SPR_STGCLEAR_SRC);
                           DDSetSpritePos(1, 100 -
                                         (SPR_STGCLEAR_000.SprSrc.Height shr 1),
                                          SPR_STGCLEAR_000);
                           SPR_STGCLEAR_ANI := 0;

                           DMPlay(SND_STAGEINI_SEG);
                           inc(STAGENO);
                           RUNSTAGE := RUNSTAGE_STGCLEAR;
                         end;
    RUNSTAGE_STGCLEAR : begin
                          if (SPR_STGCLEAR_ANI > 125) then
                            begin
                              RefreshInputs(_up, _down, _left, _right, _shot,
                                            _esc, _sF11, _sF12, _fleft, _fright);
                              if (_shot > 0) then RUNSTAGE := RUNSTAGE_STAGEBEGIN;
                            end;
                          RenderStageClear;
                        end;
    else RenderIntro;
  end;

  DDFlip(_hWnd);
end;

// ------------------------------------------------------------------

begin
  Randomize;
end.


